import bs4
import mmap
import pandas as pd


class XMLTagStream:
    '''
    Class for chunking a large XML file into separate bs4 "soup" objects based
    on a tag.
    '''
    def __init__(self, path, tag_name, with_attrs=False, encoding='utf-8'):
        self.file = open(path, encoding='utf-8')
        self.stream = mmap.mmap(
            self.file.fileno(), 0, access=mmap.ACCESS_READ
        )
        self.tag_name = tag_name
        self.encoding = encoding
        tag_end = ' ' if with_attrs else '>'
        self.start_tag = f'<{tag_name}{tag_end}'.encode(encoding)
        self.end_tag = f'</{tag_name}>'.encode(encoding)

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.stream.close()
        self.file.close()

    def __iter__(self):
        end = 0
        while (begin := self.stream.find(self.start_tag, end)) != -1:
            end = self.stream.find(self.end_tag, begin)
            yield self.parse(self.stream[begin: end + len(self.end_tag)])

    def parse(self, chunk):
        return bs4.BeautifulSoup(chunk.decode(self.encoding), features='lxml-xml')


def soup_generator(xmlfilepath, tag_name, with_attrs=False, encoding='utf-8'):
    '''
    A generator for bs4 soup objects for each tag_name in the xml file.
    :param xmlfilepath: The path to the xml file
    :param tag_name: The tag name to extract
    :param with_attrs: True to find the tag with attributes
    :param encoding: File's encoding
    :yield: A soup object for each xml tag in the file
    '''
    with XMLTagStream(
            xmlfilepath, tag_name, with_attrs=with_attrs, encoding=encoding,
    ) as stream:
        for soup in stream:
            yield soup


def html_table_scraper(
        soup_table: bs4.element.Tag,
        add_header_as_row: bool = False,
) -> pd.DataFrame:
    '''
    Scrape html table information into a dataframe.
    :param soup_table: The soup table element
    :param add_header_as_row: If True, the header row will also be added as
        a table row.
    :return: A dataframe with the scraped table data
    '''
    columns = None
    rows = list()

    def html_text(elt):
        return elt.text.replace('\xa0', ' ').strip()

    def td_text(td):
        return '\n'.join([  # add a \n between text of td elements
            x for x in [
                html_text(c)
                for c in td.children
            ]
            if x  # drop empty elements under a td
        ])

    for tr in soup_table.find_all('tr'):
        if columns is None:
            if tr.find('th') is not None:
                columns = [
                    td_text(th)
                    for th in tr.find_all('th')
                ]
                if add_header_as_row:
                    rows.append(columns)
                continue
            else:
                columns = []
        row = [
            td_text(td)
            for td in tr.find_all('td')
        ]
        if len(row) > 0:
            rows.append(row)

    return pd.DataFrame(
        rows,
        columns=columns if len(columns) > 0 else None
    )



import itertools
import json
import numpy as np
import pandas as pd
from typing import Dict, List, Set, Tuple, Union


def dicts2df(
        dicts: Union[List[Dict], List[List[Dict]]],
        rename: Dict[str, str] = None,
        item_id: str = 'item_id'
):
    '''
    Create a dataframe from a list of or list of lists of dictionaries.
    :param dicts: A list of dictionaries
    :param rename: Dictionary of {from: to} column name mappings
    :param item_id: Name of column to add with list indices
    :return: A dataframe
    '''
    dfs = [
        pd.DataFrame.from_records(rec)
        for rec in dicts
    ]
    for idx, df in enumerate(dfs):
        if rename:
            df.rename(columns=rename, inplace=True)
        if item_id:
            df[item_id] = idx
    df = pd.concat(dfs).reset_index(drop=True) if len(dfs) > 0 else pd.DataFrame()
    return df


def sort_by_strlen(
        df: pd.DataFrame,
        text_col: str,
        ascending: bool = False,
) -> pd.DataFrame:
    '''
    Sort the dataframe according to the length of the strings in a text column
    :param df: The dataframe to sort
    :param text_col: The text column to sort by
    :return: The sorted dataframe
    '''
    return df.loc[
        df[text_col].str.len().sort_values(ascending=ascending).index
    ]


def get_loc_range(bool_ser: pd.Series) -> Tuple[int, int]:
    '''
    Get the first (start index) and last True locations in the boolean series.
    :param bool_ser: A boolean series
    :return: (first_loc, last_loc) for the first and last "True" locations
    '''
    first_loc = bool_ser.index[bool_ser.argmax()]  # first True from beginning
    rev_ser = bool_ser[::-1]  # reverse
    last_loc = rev_ser.index[rev_ser.argmax()]  # first True from end
    return (first_loc, last_loc)


def explode_json_series(json_ser: pd.Series) -> pd.Series:
    '''
    Given a series with each value a json list of items, explode the series
    items.
    :param json_ser: The series with json values
    :return: The exploded series
    '''
    return json_ser[
        np.logical_and(json_ser.notna(), json_ser != '')
    ].apply(json.loads).explode()


class GroupManager:
    '''
    Class to manage groups of rows in a dataframe identified by a json list of
    numbers in a group number column such that all rows sharing the same group
    number constitute a group.
    '''

    def __init__(
            self,
            df: pd.DataFrame,
            group_num_col: str,
    ):
        '''
        Initialize with the dataframe and group number column name.
        :param df: The dataframe with rows in sorted order
        :param group_num_col: The name of the group number column
        '''
        self.idf = df     # input dataframe
        self.gcol = group_num_col
        self._cdf = self._fix_index(df)  # collapsed dataframe
        self._es = None   # expanded series
        self._glocs = None  # Dict[group_num, loc_series]
        self._mdf = None  # mask dataframe

    def _fix_index(self, df: pd.DataFrame) -> pd.DataFrame:
        # Index values must be unique and sorted in the collapsed form
        if (
                df.index.value_counts().max() > 1 or
                not (df.index.sort_values() == df.index).all()
        ):
            # Index isn't unique or in order, so need to reset the index
            df = df.reset_index(
                drop=('__orig_idx__' == None),
                allow_duplicates=False,
                names='__orig_idx__'
            )
        return df

    @property
    def df(self) -> pd.DataFrame:
        '''
        Get the collapsed dataframe such that the group number column holds the
        list of group numbers as a json string in unique rows.
        '''
        return self._cdf

    @property
    def collapsed_df(self) -> pd.DataFrame:
        '''
        Get the collapsed dataframe such that the group number column holds the
        list of group numbers as a json string in unique rows.
        '''
        return self._cdf

    @property
    def expanded_ser(self) -> pd.Series:
        '''
        Get the expanded series such that the group number column holds an
        integer and the index values are repeated for members of multiple
        groups.
        '''
        if self._es is None:
            if self._cdf is not None:
                if self.gcol in self._cdf.columns:
                    gser = self._cdf[self.gcol]
                    gser = explode_json_series(gser)
                    if len(gser) > 0:
                        self._es = gser
                if self._es is None:
                    self._es = pd.Series(
                        np.NaN,
                        index=self._cdf.index,
                        name=self.gcol,
                    )
        return self._es

    @property
    def all_group_locs(self) -> Dict[int, List[int]]:
        '''
        Get all group row locs, indexed by group_num.
        '''
        if self._glocs is None:
            edf = pd.DataFrame(self.expanded_ser)
            edf['__tmp_idx__'] = edf.index
            self._glocs = edf.groupby(
                self.gcol, group_keys=True,
            )['__tmp_idx__'].apply(list).to_dict()
        return self._glocs

    def get_group_locs(self, group_num: int) -> List[int]:
        '''
        Get the row locs for the given group, or None.
        '''
        return self.all_group_locs.get(group_num, None)

    def get_intra_ungrouped_locs(self, group_num: int) -> List[int]:
        '''
        Get the locs for rows within the given group rows, but not in the group
        '''
        result = None
        colname = f'{self.gcol}_{group_num}'
        if colname in self.mask_df:
            mcol = self.mask_df[colname]
            if mcol.any():
                startloc, endloc = get_loc_range(mcol)
                locs = ~mcol[startloc:endloc+1]
                if locs.any():
                    result = locs[locs].index.tolist()
        return result if result is not None else list()

    @property
    def grouped_locs(self) -> List[int]:
        '''
        Get all row locs that are in at least one group
        '''
        return self.mask_df.index[self.mask_df.sum(axis=1) > 0].tolist()

    @property
    def ungrouped_locs(self) -> List[int]:
        '''
        Get all row locs that aren't in any group
        '''
        return self.mask_df.index[self.mask_df.sum(axis=1) == 0].tolist()

    @property
    def all_group_nums(self) -> List[int]:
        '''
        Get the existing group numbers, or an empty list.
        '''
        result = self.expanded_ser.unique()
        result = [x for x in result if not np.isnan(x)]
        return result

    @property
    def max_group_num(self) -> int:
        '''
        Get the maximum group number present, or -1 if there are none.
        '''
        result = self.expanded_ser.max()
        return -1 if np.isnan(result) else result

    @property
    def mask_df(self) -> pd.DataFrame:
        '''
        Get a dataframe of group masks identifying rows in each group.
        '''
        if self._mdf is None:
            cdf = self.collapsed_df
            if self.gcol in cdf:
                def build_mask(gnum):
                    m = pd.Series(False, index=cdf.index)
                    m.loc[self.get_group_locs(gnum)] = True
                    return m

                self._mdf = pd.DataFrame({
                    f'{self.gcol}_{gnum}': build_mask(gnum)
                    for gnum in self.all_group_nums
                })
            else:
                self._mdf = pd.DataFrame({}, index=cdf.index)
        return self._mdf

    def mark_group(self, idx_values: pd.Series, group_num: int = None):
        '''
        Mark the rows identified by their index values as a group, either with
        the next higher available group number or with the given group number.

        If the group_num already exists, then rows without the group_num will
        be added to the group.
        
        :param idx_values: The row locs to include in the group
        :param group_num: The group number to assign, or None to auto-assign.
        '''
        df = self.collapsed_df
        if group_num is None:
            group_num = self.max_group_num + 1
        if self.gcol not in df:
            df[self.gcol] = np.NaN
        cur_values = df.loc[idx_values, self.gcol]

        def add_group(v):
            if pd.notna(v) and v != '':
                groups = set(json.loads(v))
                groups.add(group_num)
                return json.dumps(list(groups))
            else:
                return f'[{group_num}]'

        df.loc[idx_values, self.gcol] = cur_values.apply(add_group)
        self._reset_edf()

    def mark_groups(
            self,
            idx_value_lists: List[pd.Series],
            group_nums: List[int] = None,
    ):
        '''
        Convenience method to mark multiple groups at once.
        :param idx_value_lists: A list of idx_values identifying groups
        :param group_nums: A list of group numbers corresponding to each
            idx_values list. If None, then auto-increment group_nums for
            each list.
        '''
        for pos, idx_values in enumerate(idx_value_lists):
            self.mark_group(
                idx_values,
                group_num=group_nums[pos] if group_nums is not None else None
            )

    def unmark_group(self, group_num: int, idx_values: pd.Series = None):
        '''
        Remove group_num from the specified rows, or entirely.
        :param group_num: The group_num to remove
        :param idx_values: The row locs from which to remove the group_num.
            If None, then group_num will be removed from all rows in which
            it exists.
        '''
        df = self.collapsed_df
        if self.gcol in df:
            if idx_values is None:
                idx_values = df.index
            gser = df[self.gcol]
            mask = gser.mask(gser.index.isin(idx_values), False).astype(bool)
    
            def del_group(v):
                rv = v
                if pd.notna(v) and v != '':
                    groups = set(json.loads(v))
                    groups.discard(group_num)
                    rv = json.dumps(list(groups)) if len(groups) > 0 else np.NaN
                return rv

            gser.where(mask, gser.apply(del_group), inplace=True)
            self._reset_edf()

    def remove_groups(self, group_nums: List[int]):
        '''
        Convenience method to unmark (remove) the listed groups.
        :param group_num: The group numbers to remove
        '''
        for gnum in group_nums:
            self.unmark_group(gnum)

    def find_subsets(self, proper: bool = True) -> Set[int]:
        '''
        Find the groups that are subsets of other groups.
        :param proper: If True, include proper (complete) subsets.
        :return: The set of group numbers that are subsets of other groups
        '''
        rv = set()
        mdf = self.mask_df
        for g1, g2 in itertools.combinations(mdf.columns, 2):
            m1 = mdf[g1]
            m2 = mdf[g2]
            s1 = m1.sum()
            s2 = m2.sum()
            if proper or s1 != s2:
                combo_sum = (m1 & m2).sum()
                if combo_sum == s2:  # "later" smaller or equal
                    # g2 is a (proper or smaller) subset of g1
                    gn = int(g2[g2.rindex('_')+1:])
                    rv.add(gn)
                elif combo_sum == s1:  # "earlier" smaller
                    # g1 is a (smaller) subset of g2
                    gn = int(g1[g1.rindex('_')+1:])
                    rv.add(gn)
        return rv

    def remove_subsets(self, proper: bool = True):
        '''
        Convenience method to remove all groups that are subsets of others.
        :param proper: If True, include proper (complete) subsets.
        '''
        self.remove_groups(self.find_subsets(proper=proper))

    def clear_all_groups(self):
        '''
        Remove (unmark) all groups.
        '''
        df = self.collapsed_df
        if self.gcol in df.columns:
            df[self.gcol] = np.NaN
            self._reset_edf()

    def reset_group_numbers(self, start_num: int = 0):
        '''
        Reset group numbers to increase consecutively from a start.
        :param start_num: The group number to start from.
        '''
        glocs = self.all_group_locs
        self.clear_all_groups()
        for idx, (_gnum, locs) in enumerate(glocs.items()):
            self.mark_group(locs, group_num=start_num+idx)

    def _reset_edf(self):
        '''
        Reset the enhanced_df for recomputing.
        '''
        self._es = None
        self._glocs = None
        self._mdf = None

    def get_subgroup_manager(
            self,
            group_num: int,
            subgroup_num_col: str,
    ):
        '''
        Given that this group manager's groups are comprised of groups from
        another group manager, reconstruct that group manager's groups
        identified by the given group_num from this manager.

        Or, get a GroupManager for the subgroup_num_col updated to reflect only
        groups identified by this manager's group_num. Note that this returns
        a copy of this manager's dataframe with the subgroup_num_col updated
        to exclude other subgroup's groupings.

        :param group_num: This manager's group num whose subgroups to get
        :param subgroup_num_col: The name of the collapsed dataframe column
            holding subgroup values.
        :return: A new GroupManager with only the relevant subgroups marked.
        '''
        group_locs = self.get_group_locs(group_num)
        subgroup_ser = self.collapsed_df.loc[group_locs, subgroup_num_col]

        # only subgroup_locs that are not shared should remain
        es = explode_json_series(subgroup_ser)
        vc = es.index.value_counts()
        all_nums = set(es.unique())
        keeper_nums = set(es[vc[vc == 1].index])
        discard_nums = all_nums.difference(keeper_nums)

        # Make (sub) group manager with a *COPY* of this manager's df
        # so as not to destroy the subgroup column's information.
        gm = GroupManager(self.collapsed_df.copy(), subgroup_num_col)
        gm.remove_groups(discard_nums)

        return gm



import numpy as np
import os
import pandas as pd
import psycopg2
from abc import ABC, abstractmethod
from dotenv import load_dotenv
from functools import lru_cache
from typing import Any, Dict, List


class RecordFetcher(ABC):
    '''
    Abstract base class for fetching records from a data source.
    '''

    def __init__(
            self,
            id_field_name='id',
            fields_to_retrieve=None,
            one_based_ids=False
    ):
        '''
        :param id_field_name: The name of the integer "id" field for the data source.
        :param fields_to_retrieve: The subset of fields to retrieve (retrieve all if None).
        :param one_based_ids: True if the data source IDs are 1-based.
        '''
        self.id_field_name = id_field_name
        self.fields_to_retrieve = fields_to_retrieve
        self.one_based = one_based_ids

    @abstractmethod
    def get_records(
            self, ids: List[int],
            one_based: bool = False,
            fields_to_retrieve: List[str] = None
    ) -> pd.DataFrame:
        '''
        :param ids: The collection of IDs of the records to retrieve
        :param one_based: True if the ids are one-based.
        :param fields_to_retrieve: Overriding subset of self.fields_to_retrieve for this call.
        :return: A pandas dataframe with the retrieved records.
        '''
        raise NotImplementedError


class DotenvPostgresConnector:
    def __init__(self, host=None, db=None, user=None, pwd=None):
        if host is None or db is None or user is None or pwd is None:
            load_dotenv()
        
        self.host = os.getenv('POSTGRES_HOST', 'localhost') if host is None else host
        self.database = os.getenv('POSTGRES_DB', 'postgres') if db is None else db
        self.user = os.getenv('POSTGRES_USER', 'postgres') if user is None else user
        self.password = os.getenv('POSTGRES_PASSWORD', None) if pwd is None else pwd

    def get_conn(self):
        '''
        Get a connection to the database.

        :return: The database connection
        '''
        return psycopg2.connect(
            host=self.host,
            database=self.database,
            user=self.user,
            password=self.password,
        )
        

class PostgresDB():

    def __init__(self, host=None, db=None, user=None, pwd=None):
        self._connector = DotenvPostgresConnector(
            host=host, db=db, user=user, pwd=pwd
        )
        self._tables_df = None
        self._table_names = None

    @property
    def table_names(self) -> List[str]:
        '''
        Get the database table names as a list of strings.
        '''
        if self._table_names is None:
            self._table_names = self._do_get_table_names()
        return self._table_names

    @property
    def tables_df(self) -> pd.DataFrame:
        '''
        Get the database tables dataframe. Note that this is non-standard
        across db types.

        :return: The database tables dataframe
        '''
        if self._tables_df is None:
            self._tables_df = self._do_get_tables_df()
        return self._tables_df

    def get_columns(self, table_name: str) -> pd.DataFrame:
        return self.query(f'''
            SELECT *
            FROM information_schema.columns
            WHERE table_name = '{table_name}'
        ''')

    def table_head(self, table_name: str, n: int = 10) -> pd.DataFrame:
        '''
        Get a sample of N rows from the table.
        :param table_name: The table whose rows to sample
        :param n: The number of rows to return
        '''
        return self.query(f'''SELECT * FROM {table_name} LIMIT {n}''')

    def get_conn(self):
        '''
        Get a connection to the database.

        :return: The database connection
        '''
        return self._connector.get_conn()

    def _do_get_tables_df(self) -> pd.DataFrame:
        '''
        Do the work of getting the tables dataframe.
        '''
        return self.query(
            "SELECT * FROM information_schema.tables WHERE table_schema = 'public'"
        )

    def _do_get_table_names(self) -> List[str]:
        '''
        Do the work of getting table names.
        '''
        return self.tables_df['table_name'].tolist()

    @lru_cache(maxsize=2)
    def query(
            self,
            query: str,
            params: Dict[str, Any] = None,
    ) -> pd.DataFrame:
        '''
        Submit a query, returning the results as a dataframe.

        :param query: The sql query to execute
        :param params: Parameters to safely inject into the query string, where
            each parameter "param" has the form "%(param)s" in the query string
        :return: A dataframe with the results
        '''
        df = None
        with self.get_conn() as conn:
            with conn.cursor() as curs:
                if params is None:
                    curs.execute(query)
                else:
                    curs.execute(query, params)
                df = pd.DataFrame(
                    curs.fetchall(),
                    columns=[
                        desc[0] for desc in curs.description
                    ]
                )
        return df

    def execute(self, stmt: str):
        '''
        Execute a statement.

        :param stmt: The sql statement to execute
        '''
        with self.get_conn() as conn:
            with conn.cursor() as curs:
                curs.execute(stmt)

    def upload(self, table_name, df):
        '''
        Upload the dataframe data to the table.

        :param table_name: The name of the table to upload the data to
        :param df: A dataframe whose columns match the table fields that holds
                   the data to upload
        '''
        fields = ', '.join(df.columns)
        template = ', '.join(['%s'] * len(df.columns))
        if table_name not in self.table_names:
            self._create_table(table_name, df)
        with self.get_conn() as conn:
            with conn.cursor() as curs:
                sql = f'INSERT INTO {table_name} ({fields}) VALUES ' + \
                    ','.join(
                        curs.mogrify(
                            f'({template})',
                            [str(row[col]) for col in df.columns],
                        ).decode('utf-8')
                        for row in df.to_records()
                    )
                curs.execute(sql)

    def _create_table(self, table_name, df):
        '''
        Create a table for the dataframe if it doesn't already exist.
        Do not populate the table with the contents of the dataframe here.
        :param table_name: The name of the table to create
        :param df: A dataframe whose columns represent the table schema
        '''
        def psql_schema_line(df, col):
            line = None
            if np.issubdtype(df[col].dtype, np.integer):
                line = f'{col} integer'
            elif np.issubdtype(df[col].dtype, np.float64):
                line = f'{col} real'
            else:
                maxlen = max(df[col].str.len())
                line = f'{col} varchar({maxlen})'
            return line
        def build_create_table_sql(df, table_name):
            schema_lines = ','.join([
                psql_schema_line(df, col)
                for col in df.columns
            ])
            return f'CREATE TABLE IF NOT EXISTS {table_name} ({schema_lines})'
        self._tables_df = None
        self._table_names = None
        self.execute(build_create_table_sql(df, table_name))


class PostgresRecordFetcher(RecordFetcher):
    '''
    Class for fetching records from a Postgres DB.
    '''

    def __init__(
            self,
            db,
            table_name,
            id_field_name='id',
            fields_to_retrieve=None,
            one_based_ids=False
    ):
        '''
        :param db: The postgres database
        :param table_name: The name of the table of the data source.
        :param id_field_name: The name of the integer "id" field in the table.
        :param fields_to_retrieve: The subset of fields to retrieve (retrieve all if None).
        :param one_based_ids: True if the data source IDs are 1-based.
        '''
        super().__init__(
            id_field_name=id_field_name,
            fields_to_retrieve=fields_to_retrieve,
            one_based_ids=one_based_ids
        )
        self.db = db
        self.table_name = table_name

    def get_records(
            self,
            ids: List[int],
            one_based: bool = False,
            fields_to_retrieve: List[str] = None,
    ) -> pd.DataFrame:
        '''
        Get the records as a dataframe for the given IDs.
        
        :param ids: The collection of IDs of the records to retrieve
        :param one_based: True if these ids are one-based.
        :param fields_to_retrieve: Overriding subset of self.fields_to_retrieve for this call.
        :return: A pandas dataframe with the retrieved records.
        '''
        fields = '*'
        if fields_to_retrieve is None:
            fields_to_retrieve = self.fields_to_retrieve
        if fields_to_retrieve is not None:
            fields = ', '.join(fields_to_retrieve)
        offset = 0
        if one_based != self.one_based:
            offset = 1 if self.one_based else -1
        values = ', '.join([
            str(value + offset)
            for value in ids
        ])
        return self.db.query(f'''
           SELECT {fields}
           FROM {self.table_name}
           WHERE {self.id_field_name} IN ({values})
        ''')


class DictionaryRecordFetcher(RecordFetcher):
    '''
    Class for fetching records from a dictionary of IDs mapped to record values.
    '''

    def __init__(
            self,
            the_dict: Dict[int, List[Any]],
            all_field_names: List[str],
            id_field_name='id',
            fields_to_retrieve=None,
            one_based_ids=False
    ):
        '''
        :param db: The postgres database
        :param the_dict: The dictionary mapping IDs to records
        :param all_field_names: All field names in the same order as the records
        :param id_field_name: The name of the integer "id" field in the table.
        :param fields_to_retrieve: The subset of fields to retrieve (retrieve all if None).
        :param one_based_ids: True if the data source IDs are 1-based.
        '''
        super().__init__(
            id_field_name=id_field_name,
            fields_to_retrieve=fields_to_retrieve,
            one_based_ids=one_based_ids
        )
        self.the_dict = the_dict
        self.field_names = all_field_names

    def get_records(
            self,
            ids: List[int],
            one_based: bool = False,
            fields_to_retrieve: List[str] = None,
    ) -> pd.DataFrame:
        '''
        Get the records as a dataframe for the given IDs.
        
        :param ids: The collection of IDs of the records to retrieve
        :param one_based: True if these ids are one-based.
        :param fields_to_retrieve: Overriding subset of self.fields_to_retrieve for this call.
        :return: A pandas dataframe with the retrieved records.
        '''
        offset = 0
        if one_based != self.one_based:
            offset = 1 if self.one_based else -1
        records = [
            self.the_dict.get(an_id, [an_id+offset] + [None]*(len(self.field_names)-1))
            for an_id in ids
        ]
        df = pd.DataFrame(records, columns=self.field_names)
        if fields_to_retrieve is None:
            fields_to_retrieve = self.fields_to_retrieve
        if fields_to_retrieve is not None:
            df = df[fields_to_retrieve]
        return df


class DataFrameRecordFetcher(RecordFetcher):
    '''
    Class for fetching records from a pandas dataframe.
    '''

    def __init__(
            self,
            df: pd.DataFrame,
            id_field_name='id',
            fields_to_retrieve=None,
            one_based_ids=False
    ):
        '''
        :param df: The dataframe of records
        :param id_field_name: The name of the integer "id" field in the table.
        :param fields_to_retrieve: The subset of fields to retrieve (retrieve all if None).
        :param one_based_ids: True if the data source IDs are 1-based.
        '''
        super().__init__(
            id_field_name=id_field_name,
            fields_to_retrieve=fields_to_retrieve,
            one_based_ids=one_based_ids
        )
        self.df = df

    def get_records(
            self,
            ids: List[int],
            one_based: bool = False,
            fields_to_retrieve: List[str] = None,
    ) -> pd.DataFrame:
        '''
        Get the records as a dataframe for the given IDs.
        
        :param ids: The collection of IDs of the records to retrieve
        :param one_based: True if these ids are one-based.
        :param fields_to_retrieve: Overriding subset of self.fields_to_retrieve for this call.
        :return: A pandas dataframe with the retrieved records.
        '''
        offset = 0
        if one_based != self.one_based:
            offset = 1 if self.one_based else -1
        ids = [
            an_id + offset
            for an_id in ids
        ]
        df = self.df[self.df[self.id_field_name].isin(ids)]
        if fields_to_retrieve is None:
            fields_to_retrieve = self.fields_to_retrieve
        if fields_to_retrieve is not None:
            df = self.df[fields_to_retrieve]
        return df



'''
Implementation of a simple tree data structure.
'''

from __future__ import annotations
from collections import deque
from pyparsing import OneOrMore, nestedExpr
from typing import Any, Callable, List, Tuple, Union


class Tree:
    '''
    Implementation of a simple tree data structure.

    Where the tree is represented as a node containing:
      * (arbitrary) data
      * a list of (ordered) child nodes
      * a single (optional) parent node

    And each tree node is doubly linked from parent to child(ren) and from
    child to parent for efficient traversal both up (to parent) and down
    (to children) the tree.
    '''

    def __init__(
            self,
            data: Any,
            parent: Union[Tree, Any] = None,
            child_pos: int = None,
    ):
        '''
        Initialize a tree (node), optionally adding it to the given parent
        at an optional child position.
        
        :param data: The data to be contained within the node.
        :param parent: The parent node to this node.
        '''
        self._data = data
        self._children = None
        self._parent = None
        if parent is not None:
            if not isinstance(parent, Tree):
                parent = Tree(parent)
            parent.add_child(self, child_pos)

    def __repr__(self) -> str:
        '''
        :return: The string representation of this tree.
        '''
        return self.as_string(delim='  ', multiline=True)

    @property
    def data(self) -> Any:
        '''
        :return: This node's data.
        '''
        return self._data

    @data.setter
    def data(self, data: Any):
        '''
        :return: Set this node's data.
        '''
        self._data = data

    @property
    def children(self) -> List[Tree]:
        '''
        :return: This node's children -- list of child nodes.
        '''
        return self._children

    @property
    def parent(self) -> Tree:
        '''
        :return: This node's parent.
        '''
        return self._parent

    @parent.setter
    def parent(self, parent: Tree):
        '''
        :return: Set this node's parent.
        '''
        self._parent = parent

    @property
    def root(self) -> Tree:
        '''
        :return: The root of this node's tree.
        '''
        root = self
        while root.parent is not None:
            root = root.parent
        return root

    @property
    def sibnum(self) -> int:
        '''
        :return: This node's sibling number (0-based) among its parent's children
        '''
        return self._parent.children.index(self) if self._parent is not None else 0

    @property
    def num_siblings(self) -> int:
        '''
        :return: Get the number of siblings (including self) of this node
        '''
        return self._parent.num_children if self._parent is not None else 1

    @property
    def next_sibling(self) -> Tree:
        '''
        :return: This node's next sibling (or None)
        '''
        result = None
        if self._parent:
            sibs = self._parent.children
            nextsib = sibs.index(self) + 1
            if nextsib < len(sibs):
                result = sibs[nextsib]
        return result

    @property
    def prev_sibling(self) -> Tree:
        '''
        :return: This node's previous sibling (or None)
        '''
        result = None
        if self._parent:
            sibs = self._parent.children
            prevsib = sibs.index(self) - 1
            if prevsib >= 0:
                result = sibs[prevsib]
        return result

    def has_children(self) -> bool:
        '''
        :return: Whether this node has children.
        '''
        return self._children is not None and len(self._children) > 0

    @property
    def num_children(self) -> int:
        '''
        :return: The number of children under this node.
        '''
        return len(self._children) if self._children is not None else 0

    def has_parent(self) -> bool:
        '''
        :return: Whether this not has a parent.

        Note that the "root" of a tree has no parent.
        '''
        return self._parent is not None

    @property
    def depth(self) -> int:
        '''
        :return: The depth of this node in its tree.

        Where the depth is measured as the number of "hops" from the root,
        whose depth is 0, to children until this node is reached.
        '''
        result = 0
        curp = self.parent
        while curp is not None:
            curp = curp.parent
            result += 1
        return result

    def add_child(
            self,
            node_or_data: Union[Tree, Any],
            child_pos: int = None
    ) -> Tree:
        '''
        Add a child node to this node, pruning the child from any other tree.

        :param node_or_data: The node (or data for a new node) to add
        :param child_pos: The (optional) position at which to insert the node.
        :return: the (passed or new) child_node
        '''
        if self._children is None:
            self._children = []
        if isinstance(node_or_data, Tree):
            child = node_or_data
            child.prune()
            if (
                    child_pos is not None and
                    child_pos < len(self._children) and
                    child_pos >= 0
            ):
                self._children.insert(child_pos, child)
            else:
                self._children.append(child)
        else:
            child = Tree(node_or_data, self, child_pos=child_pos)
        child.parent = self
        return child

    def add_edge(
            self,
            parent_node_or_data: Union[Tree, Any],
            child_node_or_data: Union[Tree, Any],
    ) -> Tuple[Tree, Tree]:
        '''
        Add the child to the parent, using an existing (matching) child or parent.
        If the parent and child already exist, but not as parent and child, the
        child node will be moved to be a child of the parent.

        If neither the parent nor child nodes exist in this tree, the parent
        will be added as a child of this (self) node.

        :param parent_node_or_data: The parent node (or its data)
        :param child_node_or_data: The child node (or its data)
        :return: The (parent-node, child-node) tuple
        '''
        parent = None
        child = None

        if isinstance(parent_node_or_data, Tree):
            parent = parent_node_or_data
            # if it is not in this tree ...
            if len(self.find_nodes(
                lambda node: node == parent,
                include_self = True,
                only_first = True
            )) == 0:
                # ...then add it as a child of self
                self.add_child(parent)
        else:
            # can we find the data in this tree ...
            found = self.find_nodes(
                lambda node: node.data == parent_node_or_data,
                include_self = True,
                only_first = True
            )
            if len(found) > 0:
                parent = found[0]
            else:
                parent = self.add_child(parent_node_or_data)

        if isinstance(child_node_or_data, Tree):
            child = parent.add_chlld(child_node_or_data)
        else:
            # can we find the data in this tree ...
            found = self.find_nodes(
                lambda node: node.data == child_node_or_data,
                include_self = True,
                only_first = True
            )
            if len(found) > 0:
                child = parent.add_child(found[0])
            else:
                child = parent.add_child(child_node_or_data)

        return (parent, child)

    def prune(self) -> Tree:
        '''
        Prune this node from its tree.
        :return: this node's (former) parent.
        '''
        result = self._parent
        if self._parent is not None:
            if self._parent.children is not None:
                self._parent.children.remove(self)
            self._parent = None
        return result

    def find_nodes(
            self,
            accept_node_fn: Callable[[Tree], bool],
            traversal: str = 'dfs',
            include_self: bool = True,
            only_first: bool = False,
            highest_only: bool = False
    ) -> List[Tree]:
        '''
        Find nodes where accept_node_fn(tree) is True,
        using a traversal of:
          'dfs' -- depth first search
          'bfs' -- breadth first search

        :param accept_node_fn: A function returning a boolean for any Tree
            argument; True to select the node or False to skip it
        :param traversal: Either 'dfs' or 'bfs' for depth- or breadth-first
        :param include_self: True to consider this node, False to start with 
            its children
        :param only_first: True to stop after finding the first match
        :param highest_only: True to not collect any nodes under a selected node
        :return: The list of matching/accepted nodes
        '''
        queue, found = deque(), []
        if include_self:
            queue.append(self)
        elif self.children:
            queue.extend(self.children)
        while bool(queue):  # true while length(queue) > 0
            item = queue.popleft()
            if accept_node_fn(item):
                found.append(item)
                if only_first:
                    break
                elif highest_only:
                    continue
            if item.children:
                if traversal == 'dfs':
                    queue.extendleft(reversed(item.children))
                elif traversal == 'bfs':
                    queue.extend(item.children)
        return found

    def collect_terminal_nodes(
            self,
            accept_node_fn: Callable[[Tree], bool] = None,
            _found: List[Tree] = None
    ):
        '''
        Collect this tree's terminal nodes.

        :param accept_node_fn: Optional function to select which terminal nodes
            to include in the result
        :param _found: The (optional) list to which to add results
        :return: The list of collected nodes
        '''
        if _found is None:
            _found = list()
        if not self._children:
            if accept_node_fn is None or accept_node_fn(self):
                _found.append(self)
        else:
            for child in self._children:
                child.collect_terminal_nodes(accept_node_fn=accept_node_fn, _found=_found)
        return _found

    def get_edges(
            self,
            traversal: str = 'bfs',
            include_self: bool = True,
            as_data: bool = True,
    ) -> List[Tuple[Union[Tree, Any], Union[Tree, Any]]]:
        '''
        Get the edges of this tree, either as Tree nodes or data.

        :param traversal: Either 'dfs' or 'bfs' for depth- or breadth-first
        :param include_self: True to include this node, False to start with
            its children
        :param as_data: If True, then collect node data instead of Tree nodes
        :return: A list of (parent, child) tuples of edge nodes or data
        '''
        queue, result = deque(), []
        if self.children:
            queue.extend(self.children)
        while bool(queue):  # true while length(queue) > 0
            item = queue.popleft()
            if item.parent:
                if item.parent != self or include_self:
                    result.append(
                        (
                            item.parent.data, item.data
                        ) if as_data else (
                            item.parent, item
                        )
                    )
            if item.children:
                if traversal == 'dfs':
                    queue.extendleft(reversed(item.children))
                elif traversal == 'bfs':
                    queue.extend(item.children)
        return result

    def get_path(self) -> List[Tree]:
        '''
        Get the nodes from the root to this node (inclusive).
        '''
        path = deque()
        node = self
        while node is not None:
            path.appendleft(node)
            node = node.parent
        return list(path)

    def is_ancestor(self, other: Tree, self_is_ancestor: bool = False) -> bool:
        '''
        Determine whether this node is an ancestor to the other.
        :param other: The potential descendant of this node
        :param self_is_ancestor: True if this node could be considered to
            be its own ancestor
        :return: True if this node is an ancestor of the other
        '''
        result = False
        parent = other if self_is_ancestor else other.parent
        while parent is not None:
            if parent == self:
                result = True
                break
            parent = parent.parent
        return result

    def find_deepest_common_ancestor(self, other: Tree) -> Tree:
        '''
        Find the deepest common ancestor to self and other.
        :param other: The other node whose shared ancestor with self to find
        :return: The deepest common ancestor to self and other, or None
        '''
        if other is None:
            return None
        if self == other:
            return self
        result = None
        mypath, otherpath = self.get_path(), other.get_path()
        mypathlen, otherpathlen = len(mypath), len(otherpath)
        mypathidx, otherpathidx = 0, 0
        while mypathidx < mypathlen and otherpathidx < otherpathlen:
            mynode, othernode = mypath[mypathidx], otherpath[otherpathidx]
            mypathidx += 1
            otherpathidx += 1
            if mynode != othernode:
                break  # diverged
            else:
                result = mynode
        return result

    def as_string(self, delim: str = ' ', multiline: bool = False) -> str:
        '''
        Get a string representing this tree.
        :param delim: The (indentation) delimiter to use between node data
        :param multiline: True to include newlines in the result
        :param: A string representation of this tree and its descendants
        '''
        result = ''
        if self._children:
            btwn = '\n' if multiline else ''
            result = '(' + str(self.data)
            for child in self._children:
                d = (child.depth if multiline else 1) * delim
                result += btwn + d + child.as_string(delim=delim, multiline=multiline)
            result += ')'
        else:
            result = str(self.data)
        return result

    def get_deepest_left(self) -> Tree:
        '''
        :return: The terminal descendent following the left-most branches
            of this node.
        '''
        node = self
        while node is not None and node.has_children():
            node = node.children[0]
        return node

    def get_deepest_right(self) -> Tree:
        '''
        :return: The terminal descendent following the right-most branches
            of this node.
        '''
        node = self
        while node is not None and node.has_children():
            node = node.children[-1]
        return node


def build_tree_from_string(from_string: str) -> Tree:
    '''
    Build a tree object from the given tree string, e.g., output from
    the "Tree.as_string" method.
    :param from_string: The tree string
    :return: The built Tree
    '''
    if not from_string.strip().startswith('('):
        return Tree(from_string)
    data = OneOrMore(nestedExpr()).parseString(from_string)
    return build_tree_from_list(data.as_list())


def build_tree_from_list(data: Union[Any, List]) -> Tree:
    '''
    Auxiliary to build_tree for recursively building nodes from a list of
    lists.
    :param data: The tree data as a list of lists.
    :return: The root tree node
    '''
    node = None
    if isinstance(data, list) and len(data) > 0:
        node = build_tree_from_list(data[0])
        for cdata in data[1:]:
            node.add_child(build_tree_from_list(cdata))
    else:  # e.g. if isinstance(data, str):
        node = Tree(data)
    return node



import json
import os
import pandas as pd
from typing import Any, Dict, List


class RecordStore:
    '''
    Wrapper around a sequence of records represented in memory as a list of
    dictionaries and/or as a dataframe and as a tsv file on disk.
    '''

    def __init__(
            self,
            tsv_fpath: str,
            df: pd.DataFrame = None,
            sep: str = '\t',
    ):
        '''
        :param tsv_fpath: The path to the tsv file on disk. If None or
            empty, then data will not be persisted.
        :param df: An initial dataframe
        :param sep: The file separator to use (if not a tab)
        '''
        self.tsv_fpath = tsv_fpath
        self.init_df = df
        self.sep = sep
        self._df = None
        self._recs = None  # List[Dict[str, Any]]
        self._init_data(df)

    def _init_data(self, df: pd.DataFrame = None):
        '''
        Initialize store data from the tsv file.
        '''
        if self.tsv_fpath is not None and os.path.exists(self.tsv_fpath):
            self._df = pd.read_csv(self.tsv_fpath, sep=self.sep)
        else:
            self._df = df.copy() if df is not None else None
        self._recs = self._build_recs_from_df()

    def _build_recs_from_df(self) -> List[Dict[str, Any]]:
        ''' Build records from the dataframe '''
        if self._df is not None:
            recs = [
                json.loads(rec)
                for rec in self._df.to_json(
                        orient='records', lines=True
                ).strip().split('\n')
            ]
        else:
            recs = list()
        return recs

    @property
    def df(self) -> pd.DataFrame:
        ''' Get the records as a dataframe '''
        if self._df is None:
            self._df = pd.DataFrame(self._recs)
        return self._df

    @property
    def records(self) -> List[Dict[str, Any]]:
        ''' Get the records as a list of dictionaries '''
        return self._recs

    def clear(self):
        '''
        Clear the contents, starting from empty, but don't auto-"save".
        '''
        self._recs.clear()
        self._df = None

    def add_rec(self, rec: Dict[str, Any]):
        ''' Add the record '''
        self._recs.append(rec)
        self._df = None

    def save(self):
        ''' Save the records to disk as a tsv '''
        if self.tsv_fpath is not None:
            self.df.to_csv(self.tsv_fpath, sep=self.sep, index=False)

    def restore(self, df: pd.DataFrame = None):
        '''
        Restore records from the version on disk, discarding any changes.
        NOTE: If there is no backing file (e.g., tsv_fpath is None), then
        restore will discard all data and restart with the given df (if not
        None,) the init df or start anew.
        '''
        self._init_data(df if df is not None else self.init_df)



import subprocess
from typing import Callable, List


def run_command(
        handle_line_fn: Callable[[str], bool],
        command: str,
        args: List[str] = None
) -> int:
    '''
    Run a system command and do something with each line. Stop early by
    returning False from the handle_line_fn.

    Examples:

    # Print all files in the directory.
    >>> subprocess_utils.run_command(lambda x: (print(x), True)[1], 'ls -1')

    # Print files in the directory, stopping once "foo" is found.
    >>> subprocess_utils.run_command(lambda x: (print(x), x!='foo')[1], 'ls -1')

    :param handle_line_fn: A fn(output_line) that returns True to continue receiving
        lines or False to kill the process.
    :param command: A string with the command and its args or just the command
    :param args: The args for the command (if not None)
    :return: The command's return code
    '''
    the_args = command
    shell = True
    if args is not None:
        the_args = [command] + args
        shell = False
    process = subprocess.Popen(
        the_args, stdout=subprocess.PIPE, shell=shell, encoding='utf8'
    )
    while True:
        output = process.stdout.readline()
        if output == '' and process.poll() is not None:
            break
        if output:
            if not handle_line_fn(output.strip()):
                process.kill()
                break
    rc = process.poll()
    return rc



'''
Implementation of a conditional associative array (dict) using the strategy
pattern.
'''
from typing import Any, Callable, Dict


class cdict(dict):
    '''
    A dictionary that conditionally accepts attributes and/or values.

    This implementation uses the strategy pattern such that a function is
    provided on initialization for validating items that are set. If an
    attribute and/or value is not accepted during an add operation, the
    set operation will fail and the key/value will be added to the "rejected"
    property.
    '''
    
    def __init__(
            self,
            accept_fn:Callable[[Dict, Any, Any], bool],
            *args, **kwargs
    ):
        '''
        :param accept_fn: A fn(d, key, value) that returns True to accept
            the key/value into this dict d, or False to reject.
        '''
        super().__init__()
        self._rejected = dict()
        self.accept_fn = accept_fn
        #super().__init__(*args, **kwargs)
        self.update(*args, **kwargs)

    @property
    def rejected(self) -> Dict:
        return self._rejected

    def __setitem__(self, key, value) -> bool:
        if self.accept_fn(self, key, value):
            super().__setitem__(key, value)
        else:
            self._rejected[key] = value

    def setdefault(self, key, default=None):
        rv = None
        if key not in self:
            if self.accept_fn(self, key, default):
                super().__setitem__(key, default)
                rv = default
            else:
                self._rejected[key] = default
        else:
            rv = self[key]
        return rv

    def update(self, E=None, **F):
        if E is not None:
            if 'keys' in dir(E):
                for k in E:
                    self.__setitem__(k, E[k])
            else:
                for k, v in E:
                    self.__setitem__(k, v)
        if F is not None:
            for k in F:
                self.__setitem__(k, F[k])



import numpy as np
import pandas as pd
import dataknobs.structures.document as dk_doc
import dataknobs.utils.emoji_utils as emoji_utils
from abc import ABC, abstractmethod
from typing import Any, Callable, List, Tuple, Union


class CharacterFeatures(ABC):
    '''
    Class representing features of text as a dataframe with each character
    as a row and columns representing character features.
    '''
    def __init__(self, doctext: Union[dk_doc.Text, str], roll_padding: int = 0):
        '''
        :param doctext: The text to tokenize (or dk_doc.Text with its metadata)
        :param roll_padding: The number of pad characters added to each end of
            the text.
        '''
        self._doctext = doctext
        self._roll_padding = roll_padding
        self._padded_text = None

    @property
    def cdf(self) -> pd.DataFrame:
        '''
        The character dataframe with each padded text character as a row.
        '''
        raise NotImplementedError

    @property
    def doctext(self) -> dk_doc.Text:
        if isinstance(self._doctext, str):
            self._doctext = dk_doc.Text(self._doctext, None)
        return self._doctext

    @property
    def text_col(self) -> str:
        '''
        The name of the cdf column holding the text characters.
        '''
        return self.doctext.text_label

    @property
    def text(self) -> str:
        '''
        The text string.
        '''
        return self.doctext.text

    @property
    def text_id(self) -> Any:
        '''
        The ID of the text.
        '''
        return self.doctext.text_id

    @abstractmethod
    def build_first_token(
            self,
            normalize_fn: Callable[[str], str],
    ) -> 'Token':
        '''
        Build the first token as the start of tokenization.
        
        :param normalize_fn: A function to normalize a raw text term or any
          of its variations. If None, then the identity function is used.
        :return: The first text token
        '''
        raise NotImplementedError

    @property
    def roll_padding(self) -> int:
        '''
        The number of pad characters added to each end of the text.
        '''
        return self._roll_padding

    @property
    def padded_text(self) -> str:
        '''
        The text with padding included.
        '''
        if self._padded_text is None:
            padding = ' ' * self.roll_padding
            self._padded_text = padding + self.text + padding
        return self._padded_text

    def get_tokens(
            self,
            normalize_fn: Callable[[str], str] = lambda x: x,
    ) -> List['Token']:
        '''
        Get all token instances using the given normalize function.
        :param normalize_fn: The normalization function (default=identity fn)
        :return: A list of token instances
        '''
        token = self.build_first_token(normalize_fn)
        tokens = list()
        while token is not None:
            tokens.append(token)
            token = token.next_token
        return tokens        


class TextFeatures(CharacterFeatures):
    def __init__(
            self,
            doctext: Union[dk_doc.Text, str],
            split_camelcase: bool = True,
            mark_alpha: bool = False,
            mark_digit: bool = False,
            mark_upper: bool = False,
            mark_lower: bool = False,
            emoji_data: emoji_utils.EmojiData = None,
    ):
        '''
        :param doctext: The text to tokenize with its metadata
        :param split_camelcase: True to mark camel-case features
        :param mark_alpha: True to mark alpha features (separate from alnum)
        :param mark_digit: True to mark digit features (separate from alnum)
        :param mark_upper: True to mark upper features (auto-included for
            camel-case)
        :param mark_lower: True to mark lower features (auto-included for
            camel-case)
        :param emoji_data: An EmojiData instance to mark emoji BIO features

        NOTEs:
          * If emoji_data is non-null:
             * Then emojis will be treated as text (instead of as non-text)
             * If split_camelcase is True,
                * then each emoji will be in its own token
                * otherwise, each sequence of (adjacent) emojis will be treated
                  as a single token.
        '''
        # NOTE: roll_padding is determined by "roll" feature needs. Currently 1.
        super().__init__(doctext, roll_padding=1)
        self.split_camelcase = split_camelcase
        self._cdf = self._build_character_dataframe(
            split_camelcase,
            mark_alpha, mark_digit, mark_upper, mark_lower,
            emoji_data,
        )
        
    @property
    def cdf(self) -> pd.DataFrame:
        '''
        The character dataframe with each padded text character as a row.
        '''
        return self._cdf

    def build_first_token(
            self,
            normalize_fn: Callable[[str], str],
    ) -> 'Token':
        '''
        Build the first token as the start of tokenization.
        
        :param normalize_fn: A function to normalize a raw text term or any
          of its variations. If None, then the identity function is used.
        :return: The first text token
        '''
        token_mask = (
            DualTokenMask(
                self,
                self.cdf['tok_start'],
                self.cdf['tok_end'],
            )
            if self.split_camelcase
            else SimpleTokenMask(
                    self,
                    self.cdf['alnum']
            )
        )
        token = Token(token_mask, normalize_fn=normalize_fn)
        return token

    def _build_character_dataframe(
            self, split_camelcase,
            mark_alpha, mark_digit, mark_upper, mark_lower,
            emoji_data,
    ):
        if split_camelcase:
            mark_upper = True
            mark_lower = True
        cdf = pd.DataFrame({self.text_col: list(self.padded_text)})
        if mark_alpha:
            cdf['alpha'] = cdf[self.text_col].str.isalpha()
        if mark_digit:
            cdf['digit'] = cdf[self.text_col].str.isdigit()
        cdf['alnum'] = cdf[self.text_col].str.isalnum()
        cdf['space'] = cdf[self.text_col].str.isspace()
        if mark_upper:
            cdf['upper'] = cdf[self.text_col].str.isupper()
        if mark_lower:
            cdf['lower'] = cdf[self.text_col].str.islower()
        cdf['sym'] = ~(cdf['alnum'] | cdf['space'])
        if split_camelcase:
            cdf['cc1'] = np.roll(cdf['lower'], 1) & cdf['upper']
            cdf['cc2'] = (  # Mark 2nd U of UUl
                np.roll(cdf['upper'], 1) &
                cdf['upper'] &
                np.roll(cdf['lower'], -1)
            )
        # NOTE: tok_start and tok_end are both INCLUSIVE
        cdf['tok_start'] = (  # mark a char following a non-char
            cdf['alnum'] &
            ~np.roll(cdf['alnum'], 1)
        )
        cdf['tok_end'] = (  # mark a char followed by a non-char
            cdf['alnum'] &
            ~np.roll(cdf['alnum'], -1)
        )
        if split_camelcase:
            cdf['tok_start'] = (
                cdf['tok_start'] |
                cdf['cc1'] | cdf['cc2']
            )
            cdf['tok_end'] = (
                cdf['tok_end'] |
                np.roll(cdf['cc1'] | cdf['cc2'], -1)
            )
        if emoji_data is not None:
            cdf['emoji'] = pd.Series(
                list(emoji_data.emoji_bio(self.padded_text))
            )
            if split_camelcase:
                # Splitting camelcase includes splitting distinct emojis
                cdf['tok_start'] |= (cdf['emoji'] == 'B')
                cdf['tok_end'] |= (  # mark an 'I' followed by not 'I'
                    (cdf['emoji'] == 'I') &
                    np.roll(cdf['emoji'] != 'I', -1)
                )
                cdf['tok_end'] |= (  # mark an 'B' followed by not 'I'
                    (cdf['emoji'] == 'B') &
                    np.roll(cdf['emoji'] != 'I', -1)
                )
            else:
                # Not splitting camelcase keeps consecutive emojis together
                cdf['alnum'] |= (cdf['emoji'] != 'O')
        return cdf


class CharacterInputFeatures(CharacterFeatures):
    '''
    A wrapper that starts with a pre-built character features dataframe.
    '''
    def __init__(
            self,
            cdf: pd.DataFrame,
            token_mask: 'TokenMask',
            doctext: Union[dk_doc.Text, str],
            roll_padding: int = 0,
    ):
        super().__init__(doctext, roll_padding=roll_padding)
        self._cdf = cdf
        self._token_mask = token_mask
        
    @property
    def cdf(self) -> pd.DataFrame:
        '''
        The character dataframe with each padded text character as a row.
        '''
        return self._cdf

    def build_first_token(
            self,
            normalize_fn: Callable[[str], str] = None,
    ) -> 'Token':
        '''
        Build the first token as the start of tokenization.
        
        :param normalize_fn: A function to normalize a raw text term or any
          of its variations. If None, then the identity function is used.
        :return: The first text token
        '''
        token = Token(self._token_mask, normalize_fn=normalize_fn)
        return token


class TokenLoc:
    '''
    Simple structure holding information about a token's location.
    '''
    def __init__(
            self,
            start_loc: int,
            end_loc: int,
            token_num: int = None,
            start_incl: bool = True,
            end_incl: bool = False,
    ):
        '''
        Initialize with the available information.

        :param start_loc: The starting location of the token
        :param end_loc: The ending location of the token
        :param token_num: The position of the token within its text string
        :param start_incl: True if start_loc is part of the token; otherwise
            start_loc+1 is part of the token.
        :param end_incl: True if end_loc is part of the token; otherwise
            end_loc-1 is part of the token.
        '''
        self._start_loc = start_loc
        self._end_loc = end_loc
        self._token_num = token_num
        self._start_incl = int(start_incl)
        self._end_incl = int(end_incl)

    def __repr__(self) -> str:
        token_num = f'#{self._token_num}' if self._token_num >= 0 else ''
        def inclc(incl, left):
            if incl:
                return '[' if left else ']'
            else:
                return '(' if left else ')'
        return f'{token_num}{inclc(self._start_incl, True)}{self._start_loc}:{self._end_loc}{inclc(self._end_incl, False)}'

    def _incl_offset(self, wanted_incl: bool, current_incl: int) -> int:
        '''
        Get the inclusivity offset based on what is wanted versus what is.
        '''
        return int(wanted_incl) - current_incl

    @property
    def len(self) -> int:
        '''
        Get the length of the token at this location.
        '''
        return self.end_loc_excl - self.start_loc_incl

    @property
    def start_loc_incl(self) -> int:
        '''
        Get the inclusive start location.
        '''
        return self._start_loc + self._incl_offset(True, self._start_incl)

    @property
    def start_loc_excl(self) -> int:
        '''
        Get the exclusive start location.
        '''
        return self._start_loc + self._incl_offset(False, self._start_incl)

    @property
    def end_loc_incl(self) -> int:
        '''
        Get the inclusive end location.
        '''
        return self._end_loc - self._incl_offset(True, self._end_incl)

    @property
    def end_loc_excl(self) -> int:
        '''
        Get the exclusive end location.
        '''
        return self._end_loc - self._incl_offset(False, self._end_incl)

    @property
    def token_num(self) -> int:
        '''
        Get the token's position within its text string, or -1 if unknown.
        '''
        return self._token_num if self._token_num is not None else -1


class TokenMask(ABC):
    '''
    A class for accessing text characters through feature masks.
    '''
    def __init__(self, text_features: CharacterFeatures):
        self.text_features = text_features
        self.pad = self.text_features.roll_padding
        self.max_ploc = max(self.text_features.cdf.index)

    def _get_next_start(self, ref_ploc: int, token_mask: pd.Series) -> int:
        '''
        Given the end of a prior token or possible start of the next, get
        the "next" start token's starting ploc. If there is no subsequent
        token, then return None.
        :param ref_ploc: The end ploc of the prior token or start of string
        :param token_mask: The token mask to use
        :return: The ploc of the start of the next token or None.
        '''
        # if not at end of string or already at the start of a token, increment
        if ref_ploc > self.max_ploc:
            ref_ploc = None   # At end of string
        elif not token_mask.loc[ref_ploc]:
            next_ploc = increment(ref_ploc, token_mask)
            ref_ploc = next_ploc if next_ploc > ref_ploc else None
        return ref_ploc

    def get_padded_text(self, start_loc_incl:int, end_loc_excl:int) -> str:
        return self.text_features.padded_text[
            start_loc_incl:end_loc_excl
        ]

    def get_text(self, token_loc: TokenLoc) -> str:
        '''
        Get the text at the (padded) token location.
        :param token_loc: The token location
        :return: The token text
        '''
        return self.get_padded_text(
            token_loc.start_loc_incl, token_loc.end_loc_excl
        )

    @abstractmethod
    def get_next_token_loc(self, ref_ploc: int, token_num: int = -1) -> TokenLoc:
        '''
        Given the end of a prior token or possible start of the next, get
        the "next" token's location.
        If there is no subsequent token, then return None.
        :param ref_ploc: The end ploc of the prior token or start of string
        :param token_num: The token position within its text string
        :return: The TokenLoc of the next token or None.
        '''
        raise NotImplementedError

    @abstractmethod
    def get_prev_token_loc(self, from_token_loc: TokenLoc) -> TokenLoc:
        '''
        Get the previous token bounds before the given token start ploc.
        If there is no prior token, then return None.
        :param from_start: The start ploc of the token after the result
        :return: (tok_start, tok_end) plocs of the prior token or None.
        '''
        raise NotImplementedError


def increment(start_loc: int, mask: pd.Series) -> Tuple[int, bool]:
    '''
    Increment to the opposite True or False index location in the given mask
    from the given start index location.

    If the mask value at index (loc) start_idx is False, then find the
    index (loc) value where the mask is True. Then the mask values from
    start_idx (inclusive) to end_idx (exclusive) are all False.
    And vice-versa for if the mask value at start_idx is True.

    :param start_loc: The start index location
    :param mask: The boolean feature mask
    :return: end_loc Where the mask value is opposite that at start_loc
        If unable to increment (e.g., at the end of the mask or no flips),
        then end_idx will equal start_idx.
    '''
    end_loc = start_loc
    if start_loc in mask.index:
        m = mask.loc[start_loc:]
        end_iloc = m.argmin() if m.iloc[0] else m.argmax()
        if end_iloc > 0:
            end_loc = m.index[end_iloc]
    return end_loc


class SimpleTokenMask(TokenMask):
    '''
    A mask where "in" tokens are ones and "out" are zeros.
    '''
    def __init__(self, text_features: CharacterFeatures, token_mask: pd.Series):
        '''
        Initialize with the text_features and token mask.

        :param text_features: The text features to tokenize.
        :param token_mask: The token mask identifying token characters as True
            and characters between tokens as False.
        '''
        super().__init__(text_features)
        self.token_mask = token_mask
        self.revmask = token_mask[::-1]

    def get_next_token_loc(self, ref_ploc: int, token_num: int = -1) -> TokenLoc:
        '''
        Given the end of a prior token or possible start of the next, get
        the "next" token's location.
        If there is no subsequent token, then return None.
        :param ref_ploc: The end ploc of the prior token or start of string
        :param token_num: The token position within its text string
        :return: The TokenLoc of the next token or None.
        '''
        result = None
        start_ploc = self._get_next_start(ref_ploc, self.token_mask)
        if start_ploc is not None:
            end_ploc = increment(start_ploc, self.token_mask)
            result = TokenLoc(start_ploc, end_ploc, token_num=token_num)
        return result

    def get_prev_token_loc(self, from_token_loc: TokenLoc) -> TokenLoc:
        '''
        Get the previous token bounds before the given token start ploc.
        If there is no prior token, then return None.
        :param from_start: The start ploc of the token after the result
        :return: (tok_start, tok_end) plocs of the prior token or None.
        '''
        result = None

        from_loc = from_token_loc.start_loc_excl
        start_loc = increment(
            increment(from_loc, self.revmask),
            self.revmask
        )
        if start_loc != from_loc:
            start_loc += 1
            end_loc = increment(start_loc, self.token_mask)
            result = TokenLoc(
                start_loc,
                end_loc,
                token_num=from_token_loc.token_num - 1
            )
        return result


class DualTokenMask(TokenMask):
    '''
    A mask comprised of a mask for token starts and a mask for token ends.
    '''
    def __init__(
            self,
            text_features: CharacterFeatures,
            start_mask: pd.Series,
            end_mask: pd.Series,
    ):
        super().__init__(text_features)
        self.start_mask = start_mask
        self.end_mask = end_mask
        #self.tok_starts = start_mask.index[start_mask]
        #self.tok_ends = end_mask.index[end_mask]
        self.tok_starts = start_mask
        self.tok_ends = end_mask
        self.rev_starts = self.tok_starts[::-1]
        self.rev_ends = self.tok_starts[::-1]

    def _get_token_end(self, start_ploc:int) -> int:
        return self._get_next_start(start_ploc, self.tok_ends) + 1

    def get_next_token_loc(self, ref_ploc: int, token_num: int = -1) -> TokenLoc:
        '''
        Given the end of a prior token or possible start of the next, get
        the "next" token's location.
        If there is no subsequent token, then return None.
        :param ref_ploc: The end ploc of the prior token or start of string
        :param token_num: The token position within its text string
        :return: The TokenLoc of the next token or None.
        '''
        result = None
        start_ploc = self._get_next_start(ref_ploc, self.tok_starts)
        if start_ploc is not None:
            end_ploc = self._get_token_end(start_ploc)
            result = TokenLoc(start_ploc, end_ploc, token_num=token_num)
        return result

    def get_prev_token_loc(self, from_token_loc: TokenLoc) -> TokenLoc:
        '''
        Get the previous token bounds before the given token start ploc.
        If there is no prior token, then return None.
        :param from_start: The start ploc of the token after the result
        :return: (tok_start, tok_end) plocs of the prior token or None.
        '''
        result = None
        from_loc = from_token_loc.start_loc_excl
        if from_loc > self.pad:
            start_loc = increment(from_loc, self.rev_starts)
            result = TokenLoc(
                start_loc,
                self._get_token_end(start_loc),
                token_num=from_token_loc.token_num + 1
            )
        return result


class Token:
    '''
    A structure identifying the token start (inclusive) and end (exclusive)
    index positions according to text features mask.

    NOTE: The masks in CharacterFeatures include padding, which displaces indices
    relative to positions in the original text. In this class, padded indices
    are referred to with a "p".
    '''
    def __init__(
            self,
            token_mask: TokenMask,
            token_loc: TokenLoc = None,
            start_ploc: int = 0,
            prev_token: 'Token' = None,
            next_token: 'Token' = None,
            normalize_fn: Callable[[str], str] = None,
    ):
        '''
        Initialize the token pointer with text features and the token_mask.

        :param token_mask: The token mask to use.
        :param token_loc: The (padded) token location, if known or None.
            If token_loc is None and start_ploc is 0, then this will be the
            first token of the text.
        :param start_ploc: The padded character index for the start of this
            token as an alternate to specifying token_loc. If start_ploc is not
            at a token character according to the token mask, then it will be
            auto-incremented to the next token.
        :param prev_token: The token prior to this token.
        :param next_token: The token following this token.
        :param normalize_fn: A function to normalize token text.
        '''
        self.token_mask = token_mask
        self._next = next_token
        self._prev = prev_token
        self.normalize_fn = normalize_fn
        self._text = None
        self._norm_text = None
        self._pre_delims = None
        self._post_delims = None
        if token_loc is not None:
            self.token_loc = token_loc
        else:
            self.token_loc = self.token_mask.get_next_token_loc(
                max(start_ploc, token_mask.pad),
                token_num=0,
            )
        # If token_loc is None, the text is empty
        if self.token_loc is None:
            self._text = ''
            self.token_loc = TokenLoc(
                self.token_mask.max_ploc + 1,
                self.token_mask.max_ploc + 1,
                token_num=0,
            )
            self._pre_delims = ''
            self._post_delims = ''

    def __repr__(self) -> str:
        return f'Token({self.token_text}){self.token_loc}'

    @property
    def doctext(self) -> dk_doc.Text:
        '''
        Get the text object with metadata.
        '''
        return self.token_mask.text_features.doctext

    @property
    def full_text(self) -> str:
        '''
        Get the full original text of which this token is a part.
        '''
        return self.token_mask.text_features.text

    @property
    def text_id(self) -> Any:
        '''
        Get the full text ID.
        '''
        return self.token_mask.text_features.text_id

    @property
    def token_num(self) -> int:
        '''
        Get the position of this token within its text string.
        '''
        return self.token_loc.token_num

    @property
    def len(self) -> int:
        '''
        Get the length of this token.
        '''
        return self.token_loc.len

    @property
    def token_text(self) -> str:
        '''
        Get this token's original text.
        '''
        if self._text is None:
            self._text = self.token_mask.get_text(self.token_loc)
        return self._text

    @property
    def norm_text(self) -> str:
        '''
        Get this token's normalized text.
        '''
        if self._norm_text is None:
            self._norm_text = (
                self.normalize_fn(self.token_text)
                if self.normalize_fn is not None
                else self.token_text
            )
        return self._norm_text

    @property
    def start_pos(self) -> int:
        '''
        Get this token's start (incl) position in the original text.
        '''
        return self.token_loc.start_loc_incl - self.token_mask.pad

    @property
    def end_pos(self) -> int:
        '''
        Get this token's end (excl) position in the original text.
        '''
        return self.token_loc.end_loc_excl - self.token_mask.pad

    @property
    def token_pos(self) -> Tuple[int, int]:
        '''
        Get the token start (incl) and end (excl) indexes in the original text.
        '''
        return (self.start_pos, self.end_pos)

    @property
    def pre_delims(self) -> str:
        if self._pre_delims is None:
            delims = ''
            prev_loc = self.token_mask.get_prev_token_loc(self.token_loc)
            if prev_loc is not None:
                delims = self.token_mask.get_padded_text(
                    prev_loc.end_loc_excl,
                    self.token_loc.start_loc_incl
                )
            self._pre_delims = delims
        return self._pre_delims

    @property
    def post_delims(self) -> str:
        if self._post_delims is None:
            delims = ''
            next_loc  = self.token_mask.get_next_token_loc(
                self.token_loc.end_loc_excl,
            )
            if next_loc is not None:
                delims = self.token_mask.get_padded_text(
                    self.token_loc.end_loc_excl,
                    next_loc.start_loc_incl
                )
            else:
                # There isn't a next token. Get remainder of text after tok.
                delims = self.token_mask.get_padded_text(
                    self.token_loc.end_loc_excl,
                    self.token_mask.max_ploc,
                )
                
            self._post_delims = delims
        return self._post_delims

    @property
    def next_token(self) -> 'Token':
        if self._next is None:
            next_token_loc = self.token_mask.get_next_token_loc(
                self.token_loc.end_loc_excl,
                token_num=self.token_loc.token_num + 1,
            )
            if next_token_loc is not None:
                self._next = Token(
                    self.token_mask,
                    token_loc=next_token_loc,
                    prev_token=self,
                    normalize_fn=self.normalize_fn,
                )
        return self._next

    @property
    def prev_token(self) -> 'Token':
        if self._prev is None:
            prev_token_loc = self.token_mask.get_prev_token_loc(
                self.token_loc
            )
            if prev_token_loc is not None:
                self._prev = Token(
                    self.token_mask,
                    token_loc=prev_token_loc,
                    next_token=self,
                    normalize_fn=self.normalize_fn,
                )
        return self._prev

    @property
    def first_token(self) -> 'Token':
        '''
        Get the first token for this token's input.
        '''
        first = self
        while first.prev_token is not None:
            first = first.prev_token
        return first

    @property
    def last_token(self) -> 'Token':
        '''
        Get the last token for this token's input.
        '''
        last = self
        while last.next_token is not None:
            last = last.next_token
        return last



import pandas as pd
import re
import dataknobs.xization.annotations as dk_annots
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Set, Union


# Key annotation column name constants
KEY_AUTH_ID_COL = 'auth_id'


class DerivedFieldGroups(dk_annots.DerivedAnnotationColumns):
    '''
    Defines derived column types:
      * "field_type" -- The column holding they type of field of an annotation row
      * "field_group" -- The column holding the group number(s) of the field
      * "field_record" -- The column holding record number(s) of the field
    '''
    
    def __init__(
            self,
            field_type_suffix: str = '_field',
            field_group_suffix: str = '_num',
            field_record_suffix: str = '_recsnum',
    ):
        '''
        Add derived column types/names: Given an annnotation row,
          * field_type(row) == f'{row[ann_type_col]}_field'
          * field_group(row) == f'{row[ann_type_col]}_num'
          * field_record(row) == f'{row[ann_type_col])_recsnum'

        Where:
          * A field_type column holds annotation "sub"- type values, or fields
          * A field_group column identifies groups of annotation fields
          * A field_record column identifies groups of annotation field groups
        
        :param field_type_suffix: The field_type col name suffix (if not _field)
        :param field_group_suffix: The field_group col name suffix (if not _num)
        :param field_record_suffix: field_record colname sfx (if not _recsnum)
        '''
        self.field_type_suffix = field_type_suffix
        self.field_group_suffix = field_group_suffix
        self.field_record_suffix = field_record_suffix

    def get_col_value(
            self,
            metadata: dk_annots.AnnotationsMetaData,
            col_type: str,
            row: pd.Series,
            missing: str = None,
    ) -> str:
        '''
        Get the value of the column in the given row derived from col_type,
        where col_type is one of:
          * "field_type" == f"{field}_field"
          * "field_group" == f"{field}_num"
          * "field_record" == f"{field}_recsnum"

        And "field" is the row_accessor's metadata's "ann_type" col's value.

        :param metadata: The AnnotationsMetaData
        :param col_type: The type of column value to derive
        :param row: A row from which to get the value.
        :param missing: The value to return for unknown or missing column
        :return: The row value or the missing value
        '''
        value = missing
        if metadata.ann_type_col in row.index:
            field = row[metadata.ann_type_col]
            if field is not None:
                if col_type == 'field_type':
                    col_name = self.get_field_type_col(field)
                elif col_type == 'field_group':
                    col_name = self.get_field_group_col(field)
                elif col_type == 'field_record':
                    col_name = self.get_field_record_col(field)
                if col_name is not None and col_name in row.index:
                    value = row[col_name]
        return value

    def unpack_field(self, field_value: str) -> str:
        '''
        Given a field in any of its derivatives (like field type, field group
        or field record,) unpack and return the basic field value itself.
        '''
        field = field_value
        if field.endswith(self.field_record_suffix):
            field = field.replace(self.field_record_suffix, '')
        elif field.endswith(self.field_group_suffix):
            field = field.replace(self.field_group_suffix, '')
        elif field.endswith(self.field_type_suffix):
            field = field.replace(self.field_type_suffix, '')
        return field

    def get_field_name(self, field_value: str) -> str:
        '''
        Given a field name or field col name, e.g., an annotation type col's
        value (the field name); or a field type, group, or record column name,
        get the field name.
        '''
        return self.unpack_field(field_value)

    def get_field_type_col(self, field_value: str) -> str:
        '''
        Given a field name or field col name, e.g., an annotation type col's
        value; or a field type, group, or record column name, get the field
        name.
        '''
        field = self.unpack_field(field_value)
        return f'{field}{self.field_type_suffix}'

    def get_field_group_col(self, field_value: str) -> str:
        '''
        Given a field name or field col name, e.g., an annotation type col's
        value; or a field type, group, or record, get the name of the derived
        field group column.
        '''
        field = self.unpack_field(field_value)
        return f'{field}{self.field_group_suffix}'

    def get_field_record_col(self, field_value: str) -> str:
        '''
        Given a field name or field col name, e.g., an annotation type col's
        value; or a field type, group, or record, get the name of the derived
        field record column.
        '''
        field = self.unpack_field(field_value)
        return f'{field}{self.field_record_suffix}'


class AuthorityAnnotationsMetaData(dk_annots.AnnotationsMetaData):
    '''
    An extension of AnnotationsMetaData that adds an 'auth_id_col' to the
    standard (key) annotation columns (attributes).
    '''
    def __init__(
            self,
            start_pos_col: str = dk_annots.KEY_START_POS_COL,
            end_pos_col: str = dk_annots.KEY_END_POS_COL,
            text_col: str = dk_annots.KEY_TEXT_COL,
            ann_type_col: str = dk_annots.KEY_ANN_TYPE_COL,
            auth_id_col: str = KEY_AUTH_ID_COL,
            sort_fields: List[str] = (dk_annots.KEY_START_POS_COL, dk_annots.KEY_END_POS_COL),
            sort_fields_ascending: List[bool] = (True, False),
            **kwargs
    ):
        '''
        Initialize with key (and more) column names and info.

        Key column types:
          * start_pos
          * end_pos
          * text
          * ann_type
          * auth_id

        NOTEs:
          * Actual table columns can be named arbitrarily
             * BUT: interactions through annotations classes and interfaces
               relating to the "key" columns must use the key column constants

        :param start_pos_col: Col name for the token starting position
        :param end_pos_col: Col name for the token ending position
        :param text_col: Col name for the token text
        :param ann_type_col: Col name for the annotation types
        :param auth_id_col: Col name for the authority value ID
        :param sort_fields: The col types relevant for sorting annotation rows
        :param sort_fields_ascending: To specify sort order of sort_fields
        :param **kwargs: More column types mapped to column names
        '''
        super().__init__(
            start_pos_col=start_pos_col,
            end_pos_col=end_pos_col,
            text_col=text_col,
            ann_type_col=ann_type_col,
            sort_fields=sort_fields,
            sort_fields_ascending=sort_fields_ascending,
            auth_id=auth_id_col,
            **kwargs,
        )

    @property
    def auth_id_col(self) -> str:
        ''' Get the column name for the auth_id '''
        return self.data[KEY_AUTH_ID_COL]


class AuthorityAnnotationsBuilder(dk_annots.AnnotationsBuilder):
    '''
    An extension of an AnnotationsBuilder that adds the 'auth_id' column.
    '''

    def __init__(
            self,
            metadata: AuthorityAnnotationsMetaData=None,
            data_defaults: Dict[str, Any]=None,
    ):
        '''
        :param metadata: The authority annotations metadata
        :param data_defaults: Dict[ann_colname, default_value] with default
            values for annotation columns
        '''
        super().__init__(
            metadata if metadata is not None else AuthorityAnnotationsMetaData(),
            data_defaults
        )

    def build_annotation_row(
            self,
            start_pos: int,
            end_pos: int,
            text: str,
            ann_type: str,
            auth_id: str,
            **kwargs
    ) -> Dict[str, Any]:
        '''
        Build an annotation row with the mandatory key values and those from
        the remaining keyword arguments.

        For those kwargs whose names match metadata column names, override the
        data_defaults and add remaining data_default attributes.

        :param result_row_dict: The result row dictionary being built
        :param start_pos: The token start position
        :param end_pos: The token end position
        :param text: The token text
        :param ann_type: The annotation type
        :param auth_id: The authority ID for the row
        :return: The result_row_dict
        '''
        return self.do_build_row({
            self.metadata.start_pos_col: start_pos,
            self.metadata.end_pos_col: end_pos,
            self.metadata.text_col: text,
            self.metadata.ann_type_col: ann_type,
            self.metadata.auth_id_col: auth_id,
        }, **kwargs)


class AuthorityData:
    '''
    A wrapper for authority data.
    '''
    def __init__(self, df: pd.DataFrame, name: str):
        self._df = df
        self.name = name

    @property
    def df(self) -> pd.DataFrame:
        '''
        Get the authority data in a dataframe
        '''
        return self._df

    def lookup_values(self, value: Any, is_id=False) -> pd.DataFrame:
        '''
        Lookup authority value(s) for the given value or value id.
        :param value: A value or value_id for this authority
        :param is_id: True if value is an ID
        :return: The applicable authority dataframe rows.
        '''
        col = self.df.index if is_id else self.df[self.name]
        return self.df[col == value]


class Authority(dk_annots.Annotator):
    '''
    A class for managing and defining tabular authoritative data for e.g.,
    taxonomies, etc., and using them to annotate instances within text.
    '''
    def __init__(
            self,
            name:str,
            auth_anns_builder: AuthorityAnnotationsBuilder = None,
            authdata: AuthorityData = None,
            field_groups: DerivedFieldGroups = None,
            anns_validator: Callable[['Authority', Dict[str,Any]], bool] = None,
            parent_auth:'Authority' = None,
    ):
        '''
        Initialize with this authority's metadata.
        :param name: This authority's entity name
        :param auth_anns_builder: The authority annotations row builder to use
            for building annotation rows.
        :param authdata: The authority data
        :param field_groups: The derived field groups to use
        :param anns_validator: fn(auth, anns_dict_list) that returns True if
           the list of annotation row dicts are valid to be added as
           annotations for a single match or "entity".
        :param parent_auth: This authority's parent authority (if any)
        '''
        super().__init__(name)
        self.anns_builder = (
            auth_anns_builder if auth_anns_builder is not None
            else AuthorityAnnotationsBuilder()
        )
        self.authdata = authdata
        self.field_groups = (
            field_groups if field_groups is not None
            else DerivedFieldGroups()
        )
        self.anns_validator = anns_validator
        self._parent = parent_auth

    @property
    def metadata(self) -> AuthorityAnnotationsMetaData:
        ''' Get the meta-data '''
        return self.anns_builder.metadata

    @property
    def parent(self) -> 'Authority':
        '''
        Get this authority's parent, or None.
        '''
        return self._parent

    @abstractmethod
    def has_value(self, value: Any) -> bool:
        '''
        Determine whether the given value is in this authority.
        :param value: A possible authority value.
        :return: True if the value is a valid entity value.
        '''
        raise NotImplementedError

    def annotate_input(
            self,
            text_obj: Union[dk_annots.AnnotatedText, str],
            **kwargs,
    ) -> dk_annots.Annotations:
        '''
        Find and annotate this authority's entities in the document text
        as dictionaries like:
        [
            {
                'input_id': <id>,
                'start_pos': <start_char_pos>,
                'end_pos': <end_char_pos>,
                'entity_text': <entity_text>,
                'ann_type': <authority_name>,
                '<auth_id>': <auth_value_id_or_canonical_form>,
                'confidence': <confidence_if_available>,
            },
        ]
        :param text_obj: The text object or string to process.
        :return: An Annotations instance
        '''
        if text_obj is not None:
            if isinstance(text_obj, str) and len(text_obj.strip()) > 0:
                text_obj = dk_annots.AnnotatedText(
                    text_obj,
                    annots_metadata = self.metadata,
                )
        if text_obj is not None:
            annotations = self.add_annotations(text_obj)
        return annotations

    @abstractmethod
    def add_annotations(
            self,
            text_obj: dk_annots.AnnotatedText,
    ) -> dk_annots.Annotations:
        '''
        Method to do the work of finding, validating, and adding annotations.
        :param text_obj: The annotated text object to process and add annotations.
        :return: The added Annotations
        '''
        raise NotImplementedError

    def validate_ann_dicts(self, ann_dicts: List[Dict[str, Any]]) -> bool:
        '''
        The annotation row dictionaries are valid if:
          * They are non-empty
          * and
             * either there is no annotations validator
             * or they are valid according to the validator
        :param ann_dicts: Annotation dictionaries
        :return: True if valid
        '''
        return (
                len(ann_dicts) > 0 and (
                    self.anns_validator is None or
                    self.anns_validator(self, ann_dicts)
                )
        )

    def compose(
            self,
            annotations: dk_annots.Annotations,
    ) -> dk_annots.Annotations:
        '''
        Compose annotations into groups.
        :param annotations: The annotations
        :return: composed annotations
        '''
        return annotations

    def build_annotation(
            self,
            start_pos: int = None,
            end_pos: int = None,
            entity_text: str = None,
            auth_value_id: Any = None,
            conf: float = 1.0,
            **kwargs,
    ) -> Dict[str, Any]:
        '''
        Build annotations with the given components.
        '''
        return self.anns_builder.build_annotation_row(
            start_pos, end_pos, entity_text, self.name, auth_value_id,
            auth_valconf=conf, **kwargs
        )


class AnnotationsValidator(ABC):
    '''
    A base class with helper functions for performing validations on annotation
    rows.
    '''

    def __call__(
            self,
            auth: Authority,
            ann_row_dicts: List[Dict[str, Any]],
    ) -> bool:
        '''
        Call function to enable instances of this type of class to be passed in
        as a anns_validator function to an Authority.
        :param auth: The authority proposing annotations
        :param ann_row_dicts: The proposed annotations
        :return: True if the annotations are valid; otherwise, False
        '''
        return self.validate_annotation_rows(
            AnnotationsValidator.AuthAnnotations(auth, ann_row_dicts)
        )

    @abstractmethod
    def validate_annotation_rows(
            self,
            auth_annotations: 'AnnotationsValidator.AuthAnnotations',
    ) -> bool:
        '''
        Determine whether the proposed authority annotation rows are valid.
        :param auth_annotations: The AuthAnnotations instance with the
            proposed data.
        :return: True if valid; False if not.
        '''
        raise NotImplementedError

    class AuthAnnotations:
        '''
        A wrapper class for convenient access to the entity annotations.
        '''
        def __init__(self, auth: Authority, ann_row_dicts: List[Dict[str, Any]]):
            self.auth = auth
            self.ann_row_dicts = ann_row_dicts
            self._row_accessor = None  # AnnotationsRowAccessor
            self._anns = None  # Annotations
            self._atts = None  # Dict[str, str]
    
        @property
        def row_accessor(self) -> dk_annots.AnnotationsRowAccessor:
            '''
            Get the row accessor for this instance's annotations.
            '''
            if self._row_accessor is None:
                self._row_accessor = dk_annots.AnnotationsRowAccessor(
                    self.auth.metadata, derived_cols=self.auth.field_groups
                )
            return self._row_accessor
    
        @property
        def anns(self) -> dk_annots.Annotations:
            ''' Get this instance's annotation rows as an annotations object '''
            if self._anns is None:
                self._anns = dk_annots.Annotations(self.auth.metadata)
                for row_dict in self.ann_row_dicts:
                    self._anns.add_dict(row_dict)
            return self._anns
    
        @property
        def df(self) -> pd.DataFrame:
            ''' Get the annotation's dataframe '''
            return self.anns.df
    
        def get_field_type(self, row: pd.Series) -> str:
            ''' Get the entity field type value '''
            return self.row_accessor.get_col_value('field_type', row, None)
    
        def get_text(self, row: pd.Series) -> str:
            ''' Get the entity text from the row '''
            return self.row_accessor.get_col_value(
                self.auth.metadata.text_col, row, None
            )

        @property
        def attributes(self) -> Dict[str, str]:
            ''' Get this instance's annotation entity attributes '''
            if self._atts is None:
                self._atts = {
                    self.get_field_type(row): self.get_text(row)
                    for _, row in self.df.iterrows()
                }
            return self._atts
    
        def colval(self, col_name, row) -> Any:
            ''' Get the column's value from the given row '''
            return self.row_accessor.get_col_value(col_name, row)


class AuthorityFactory(ABC):
    '''
    A factory class for building an authority.
    '''
    @abstractmethod
    def build_authority(
            self,
            name: str,
            auth_anns_builder: AuthorityAnnotationsBuilder,
            authdata: AuthorityData,
            parent_auth: Authority = None,
    ) -> Authority:
        '''
        Build an authority with the given name and data.
        :param name: The authority name
        :param auth_anns_builder: The authority annotations row builder to use
            for building annotation rows.
        :param authdata: The authority data
        :param parent_auth: The parent authority.
        :return: The authority
        '''
        raise NotImplementedError


class LexicalAuthority(Authority):
    '''
    A class for managing named entities by ID with associated values and
    variations.
    '''
    def __init__(
            self,
            name:str,
            auth_anns_builder: AuthorityAnnotationsBuilder = None,
            authdata: AuthorityData = None,
            field_groups: DerivedFieldGroups = None,
            anns_validator: Callable[['Authority', Dict[str,Any]], bool] = None,
            parent_auth:'Authority' = None,
    ):
        '''
        Initialize with this authority's metadata.
        :param name: This authority's entity name
        :param auth_anns_builder: The authority annotations row builder to use
            for building annotation rows.
        :param authdata: The authority data
        :param field_groups: The derived field groups to use
        :param anns_validator: fn(auth, anns_dict_list) that returns True if
           the list of annotation row dicts are valid to be added as
           annotations for a single match or "entity".
        :param parent_auth: This authority's parent authority (if any)
        '''
        super().__init__(
            name,
            auth_anns_builder=auth_anns_builder,
            authdata=authdata,
            field_groups=field_groups,
            anns_validator=anns_validator,
            parent_auth=parent_auth,
        )

    @abstractmethod
    def get_value_ids(self, value:Any) -> Set[Any]:
        '''
        Get all IDs associated with the given value. Note that typically
        there is a single ID for any value, but this allows for inherent
        ambiguities in the authority.
        :param value: An authority value
        :return: The associated IDs or an empty set if the value is not valid.
        '''
        raise NotImplementedError

    @abstractmethod
    def get_values_by_id(self, value_id:Any) -> Set[Any]:
        '''
        Get all values for the associated value ID. Note that typically
        there is a single value for an ID, but this allows for inherent
        ambiguities in the authority.
        
        :param value: An authority value
        :return: The associated IDs or an empty set if the value is not valid.
        '''
        raise NotImplementedError

    @abstractmethod
    def get_id_by_variation(self, variation:str) -> Set[str]:
        '''
        Get the IDs of the value(s) associated with the given variation.
        :param variation: Variation text
        :return: The possibly empty set of associated value IDS.
        '''
        raise NotImplementedError

    @abstractmethod
    def find_variations(
            self,
            variation: str,
            starts_with: bool = False,
            ends_with: bool =False,
            scope: str = 'fullmatch',
    ) -> pd.Series:
        '''
        Find all matches to the given variation.
        :param variation: The text to find; treated as a regular expression
            unless either starts_with or ends_with is True.
        :param starts_with: When True, find all terms that start with the
            variation text.
        :param ends_with: When True, find all terms that end with the variation
            text.
        :param scope: 'fullmatch' (default), 'match', or 'contains' for
            strict, less strict, and least strict matching
        :param category_constraints: When present, limit results to terms with
            the given constraints.
        :return: The matching variations as a pd.Series

        Note only the first true of starts_with, ends_with, and scope will
        be applied. If none of these are true, an full match on the pattern
        is performed.
        '''
        raise NotImplementedError


class RegexAuthority(Authority):
    '''
    A class for managing named entities by ID with associated values and
    variations.
    '''
    def __init__(
            self,
            name:str,
            regex:re.Pattern,
            canonical_fn: Callable[[str, str], Any] = None,
            auth_anns_builder: AuthorityAnnotationsBuilder = None,
            authdata: AuthorityData = None,
            field_groups: DerivedFieldGroups = None,
            anns_validator: Callable[[Authority, Dict[str,Any]], bool] = None,
            parent_auth:'Authority' = None,
    ):
        '''
        Initialize with this authority's entity name.
        :param name: The authority name
        :param regex: The regular expression to apply
        :param canonical_fn: A function, fn(match_text, group_name), to
            transform input matches to a canonical form as a value_id.
            Where group_name will be None and the full match text will be
            passed in if there are no group names. Note that the canonical form
            is computed before the match_validator is applied and its value
            will be found as the value to the <auth_id> key.
        :param auth_anns_builder: The authority annotations row builder to use
            for building annotation rows.
        :param authdata: The authority data
        :param field_groups: The derived field groups to use
        :param anns_validator: A validation function for each regex match
            formed as a list of annotation row dictionaries, one row dictionary
            for each matching regex group. If the validator returns False,
            then the annotation rows will be rejected. The entity_text key
            will hold matched text and the <auth_name>_field key will hold
            the group name or number (if there are groups with or without names)
            or the <auth_name> if there are no groups in the regular expression.
            Note that the validator function takes the regex authority instance
            as its first parameter to provide access to the field_groups, etc.
            The validation_fn signature is: fn(regexAuthority, ann_row_dicts)
            and returns a boolean.
        :param parent_auth: This authority's parent authority (if any)
        :param group_name_colname: The name of the annotations column for
            the regex group names, or None to ignore group_names.

        NOTE: If the regular expression has capturing groups, each group
        will result in a separate entity, with the group name if provided
        in the regular expression as ...(?P<group_name>group_regex)...
        '''
        super().__init__(
            name,
            auth_anns_builder=auth_anns_builder,
            authdata=authdata,
            field_groups=field_groups,
            anns_validator=anns_validator,
            parent_auth=parent_auth,
        )
        self.regex = regex
        self.canonical_fn = canonical_fn

    def has_value(self, value: Any) -> re.Match:
        '''
        Determine whether the given value is in this authority.
        :param value: A possible authority value.
        :return: None if the value is not a valid entity value; otherwise,
            return the re.Match object.
        '''
        return self.regex.match(str(value))

    def add_annotations(
            self,
            text_obj: dk_annots.AnnotatedText,
    ) -> dk_annots.Annotations:
        '''
        Method to do the work of finding, validating, and adding annotations.
        :param text_obj: The annotated text object to process and add annotations.
        :return: The added Annotations
        '''
        for match in re.finditer(self.regex, text_obj.text):
            ann_dicts = list()
            if match.lastindex is not None:
                if len(self.regex.groupindex) > 0:  # we have named groups
                    for group_name, group_num in self.regex.groupindex.items():
                        group_text = match.group(group_num)
                        kwargs = {
                            self.field_groups.get_field_type_col(self.name): group_name
                        }
                        ann_dicts.append(self.build_annotation(
                            start_pos=match.start(group_name),
                            end_pos=match.end(group_name),
                            entity_text=group_text,
                            auth_value_id=self.get_canonical_form(group_text, group_name),
                            **kwargs
                        ))
                else:  # we have only numbers for groups
                    for group_num, group_text in enumerate(match.groups()):
                        group_num += 1
                        kwargs = {
                            self.field_groups.get_field_type_col(self.name): group_num
                        }
                        ann_dicts.append(self.build_annotation(
                            start_pos=match.start(group_num),
                            end_pos=match.end(group_num),
                            entity_text=group_text,
                            auth_value_id=self.get_canonical_form(group_text, group_num),
                            **kwargs
                        ))
            else:  # we have no groups
                ann_dicts.append(self.build_annotation(
                    start_pos=match.start(),
                    end_pos=match.end(),
                    entity_text=match.group(),
                    auth_value_id=self.get_canonical_form(match.group(), self.name),
                ))
            if self.validate_ann_dicts(ann_dicts):
                # Add non-empty, valid annotation dicts to the result
                text_obj.annotations.add_dicts(ann_dicts)
        return text_obj.annotations

    def get_canonical_form(self, entity_text:str, entity_type:str) -> Any:
        if self.canonical_fn is not None:
            entity_text = self.canonical_fn(entity_text, entity_type)
        return entity_text


class AuthoritiesBundle(Authority):
    '''
    An authority for expressing values through multiple bundled "authorities"
    like dictionary-based and/or multiple regular expression patterns.
    '''

    def __init__(
            self,
            name:str,
            auth_anns_builder: AuthorityAnnotationsBuilder = None,
            authdata: AuthorityData = None,
            field_groups: DerivedFieldGroups = None,
            parent_auth:'Authority' = None,
            anns_validator: Callable[['Authority', Dict[str,Any]], bool] = None,
            auths: List[Authority] = None,
    ):
        '''
        :param name: This authority's entity name
        :param auth_anns_builder: The authority annotations row builder to use
            for building annotation rows.
        :param authdata: The authority data
        :param field_groups: The derived field groups to use
        :param anns_validator: fn(auth, anns_dict_list) that returns True if
           the list of annotation row dicts are valid to be added as
           annotations for a single match or "entity".
        :param parent_auth: This authority's parent authority (if any)
        :param auths: The authorities to bundle together.
        '''
        super().__init__(
            name,
            auth_anns_builder=auth_anns_builder,
            authdata=authdata,
            field_groups=field_groups,
            anns_validator=anns_validator,
            parent_auth=parent_auth,
        )
        self.auths = auths.copy() if auths is not None else list()

    def add(self, auth: Authority):
        '''
        Add the authority to this bundle
        :param auth: The authority to add.
        '''
        self.auths.append(auth)

    def has_value(self, value: Any) -> bool:
        '''
        Determine whether the given value is in this authority.
        :param value: A possible authority value.
        :return: True if the value is a valid entity value.
        '''
        for auth in self.auths:
            if auth.has_value(value):
                return True
        return False

    def add_annotations(
            self,
            text_obj: dk_annots.AnnotatedText,
    ) -> dk_annots.Annotations:
        '''
        Method to do the work of finding, validating, and adding annotations.
        :param text_obj: The annotated text object to process and add annotations.
        :return: The added Annotations
        '''
        for auth in self.auths:
            auth.annotate_input(text_obj)
        return text_obj.annotations



from abc import ABC
from typing import Any, Dict


# Key text metadata attributes
TEXT_ID_ATTR = 'text_id'
TEXT_LABEL_ATTR = 'text_label'
TEXT_LABEL = 'text'


class MetaData(ABC):
    '''
    Container for managing and providing access to meta-data.
    '''

    def __init__(self, key_data: Dict[str, Any], **kwargs):
        '''
        Initialize with the mandatory or "key" data and any additional optional
        values.
        '''
        self._data = key_data.copy() if key_data is not None else dict()
        if kwargs is not None:
            self._data.update(kwargs)

    @property
    def data(self) -> Dict[str, Any]:
        return self._data

    def get_value(self, attribute: str, missing: str = None) -> Any:
        '''
        Get the value for the given attribute, or the "missing" value.
        :param attribute: The meta-data attribute whose value to get
        :param missing: The missing value
        :return: The attribute's value or the missing value.
        '''
        return self.data.get(attribute, missing)


class TextMetaData(MetaData):
    '''
    Container for text meta-data
    '''

    def __init__(
            self,
            text_id: Any,
            text_label: str = TEXT_LABEL,
            **kwargs
    ):
        super().__init__(
            {
                TEXT_ID_ATTR: text_id,
                TEXT_LABEL_ATTR: text_label,
            },
            **kwargs
        )

    @property
    def text_id(self) -> Any:
        return self.data[TEXT_ID_ATTR]

    @property
    def text_label(self) -> str:
        return self.data[TEXT_LABEL_ATTR]


class Text:
    '''
    Wrapper for a text string for analysis.
    '''

    def __init__(
            self,
            text: str,
            metadata: TextMetaData,
    ):
        self._text = text
        self._metadata = metadata

    @property
    def text(self) -> str:
        return self._text

    @property
    def text_id(self) -> Any:
        return self.metadata.text_id

    @property
    def text_label(self) -> str:
        return self.metadata.text_label

    @property
    def metadata(self) -> str:
        if self._metadata is None:
            self._metadata = TextMetaData(text_id=0)
        return self._metadata



'''
Utilities for working with unicode emojis.


References:

* https://home.unicode.org/emoji/about-emoji/
   * https://unicode.org/emoji/techindex.html
      * https://www.unicode.org/Public/emoji/15.0/
         * https://www.unicode.org/Public/emoji/15.0/emoji-test.txt


Emoji basics:

* Emojis are represented by one ore more characters
* Compound emojis are built using a "zero width joiner" of U+200D
   * This distinguishes a complex compound emoji from an adjacent emoji
   * That can be displayed as a single emoji
   * Such sequences are called "ZWJ" sequences
* An optional "variation selector 16", U+FE0F, can follow an emoji's chars
   * This indicates to render the emoji with its variation(s)

Version and Updates:

* Current unicode version is 15.0
   * Data collection is from emoji-test.txt
      * Watch out for format changes in that file if/when updating the version

Usage:

* Download the desired version's emoji-test.txt resource.
* Create an EmojiData instance with the path to the resource
* Use EmojiData to
   * Mark emoji locations (BIO) in text
   * Extract emojis from text
   * Lookup, browse, investigate emojis with their metadata (Emoji dataclass)
'''
import os
import re
from collections import Counter, defaultdict
from dataclasses import dataclass
from typing import Dict, List, Set


LATEST_EMOJI_DATA = 'resources/emoji-test.15.0.txt'


ETESTLINE_RE = re.compile(r'^([0-9A-F ]+); ([\w\-]+)\s+#.+(E\d+\.\d+) (.*)$')
STATUS_COUNT_RE = re.compile(r'^# ([\w\-]+)\s+:\s+(\d+).*$')


ZERO_WIDTH_JOINER = chr(int('0x200D', 16))
VARIATION_SELECTOR_16 = chr(int('0xFE0F', 16))

SPECIAL_CHARS = {
    ZERO_WIDTH_JOINER,
    VARIATION_SELECTOR_16,
}


@dataclass
class Emoji:
    emoji: str            # The emoji characters
    status: str           # component  (fully-  minimally-  un) + qualified
    since_version: str    # "Ex.x"
    short_name: str       # (english) short name
    group: str = None     # test file "group" name
    subgroup: str = None  # test file "subgroup" name


def build_emoji_dataclass(emoji_test_line: str) -> Emoji:
    '''
    Build an Emoji dataclass from the emoji-test file line.
    '''
    result = None
    m = ETESTLINE_RE.match(emoji_test_line)
    if m:
        result = Emoji(
            ''.join(chr(int(x, 16)) for x in m.group(1).split()),
            m.group(2),
            m.group(3),
            m.group(4).strip()
        )
    return result


def get_emoji_seq(emoji_str: str, as_hex: bool = False) -> List[int]:
    return [
        hex(ord(x)) for x in emoji_str
    ] if as_hex else [
        ord(x) for x in emoji_str
    ]


class EmojiData:
    '''
    Class for interpreting the unicode emoji_test.txt file.
    '''

    def __init__(self, emoji_test_path: str):
        self.emojis = dict()  # emojichars -> EmojiData
        self._echars = None
        self._ldepechars = None
        self._rdepechars = None
        self._load_emoji_test(emoji_test_path)

    @property
    def echars(self) -> List[int]:
        '''
        Code points that alone are an emoji code point.
        '''
        if self._echars is None:
            self._compute_echars()
        return self._echars

    @property
    def ldepechars(self) -> Dict[int, Set[int]]:
        '''
        Code points are not an emoji alone, but precede emoji code points,
        where ldepechars[cp] is the set of all emoji code points that follow cp.
        '''
        if self._ldepechars is None:
            self._compute_echars()
        return self._ldepechars

    @property
    def rdepechars(self) -> Dict[int, Set[int]]:
        '''
        Code points are not an emoji alone, but follow emoji code points,
        where rdepechars[cp] is the set of all emoji code points that precede cp.
        '''
        if self._rdepechars is None:
            self._compute_echars()
        return self._rdepechars

    def _compute_echars(self):
        loneechars = [
            emoji[0]
            for emoji in self.emojis
            if len(emoji) == 1
        ]
        ldepechars = defaultdict(set)
        rdepechars = defaultdict(set)
        for seq in self.emojis:
            for idx, echar in enumerate(seq):
                if (echar not in loneechars and echar not in SPECIAL_CHARS):
                    if idx + 1 < len(seq):
                        ldepechars[echar].add(seq[idx+1])
                    else:
                        rdepechars[echar].add(seq[idx-1])
        self._echars = loneechars
        self._ldepechars = ldepechars
        self._rdepechars = rdepechars

    def emojis_with_cp(self, cp: int) -> List[Emoji]:
        '''
        Find emojis containing the given code point.
        :param cp: The code point
        :return: The list of emoji dataclasses
        '''
        return [
            e for emoji, e in self.emojis.items()
            if cp in get_emoji_seq(emoji)
        ]

    def emoji_bio(self, emoji_text: str) -> str:
        '''
        Given a string of text, create a "BIO" string to identify Begin,
        Internal, and Outer emoji characters in the text.
        :param text: The input text
        :return: a BIO string
        '''
        result = list()
        start_pos = -1
        textlen = len(emoji_text)
        prevc = None
        for idx, c in enumerate(emoji_text):
            isechar = c in self.echars
            isrechar = c in self.rdepechars and prevc in self.rdepechars[c]
            islechar = (
                idx+1 < textlen and
                c in self.ldepechars and
                emoji_text[idx+1] in self.ldepechars[c]
            )
            issp = c in SPECIAL_CHARS

            if start_pos < 0:
                if isechar or islechar:
                    start_pos = idx
                    result.append('B')
                else:
                    result.append('O')
            else:
                if not isrechar and not issp:
                    if isechar or islechar:
                        if prevc != ZERO_WIDTH_JOINER:
                            start_pos = idx
                            result.append('B')
                        else:
                            result.append('I')
                    else:
                        start_pos = -1
                        result.append('O')
                else:
                    result.append('I')
            prevc = c
        return ''.join(result)

    def get_emojis(self, text: str) -> List['EmojiData']:
        '''
        Get emoji data for all emojis in the given text.
        :param text: Arbitrary text
        :return: The (possibly empty) list of emojis found
        '''
        result = list()
        bio = self.emoji_bio(text)
        biolen = len(bio)
        start_pos = 0
        while 'B' in bio[start_pos:]:
            start_pos = bio.index('B', start_pos)
            end_pos = start_pos + 1
            while end_pos < biolen and bio[end_pos] == 'I':
                end_pos += 1
            result.append(self.emojis[text[start_pos:end_pos]])
            start_pos = end_pos
        return result

    def _load_emoji_test(self, emoji_test_path: str):
        curgroup = ''
        cursubgroup = ''
        with open(emoji_test_path, 'r', encoding='utf-8') as f:
            for line in f:
                if line.startswith('# group:'):
                    curgroup = line[9:].strip()
                elif line.startswith('# subgroup:'):
                    cursubgroup = line[12:].strip()
                elif line.startswith('# Status Counts'):
                    # Verify actual counts
                    c = Counter()
                    for e in self.emojis.values():
                        c[e.status] += 1
                    # with expectations
                    for l in f:
                        if l.startswith('# '):
                            m = STATUS_COUNT_RE.match(l)
                            if m:
                                assert c[m.group(1)] == int(m.group(2))
                        else:
                            break
                elif not line.startswith('#'):
                    line = line.strip()
                    if line:
                        e = build_emoji_dataclass(line)
                        if e:
                            e.group = curgroup
                            e.subgroup = cursubgroup
                            self.emojis[e.emoji] = e


def load_emoji_data() -> EmojiData:
    '''
    Load latest emoji-test.txt or reference EMOJI_TEST_DATA env var.
    '''
    result = None
    datapath = os.environ.get('EMOJI_TEST_DATA', LATEST_EMOJI_DATA)
    if os.path.exists(datapath):
        result = EmojiData(datapath)
    return result



import more_itertools
import numpy as np
import pandas as pd
import dataknobs.xization.annotations as dk_anns
import dataknobs.xization.authorities as dk_auth
import dataknobs.structures.document as dk_doc
import dataknobs.xization.masking_tokenizer as dk_tok
import dataknobs.utils.emoji_utils as emoji_utils
from abc import abstractmethod
from collections import defaultdict
from typing import Any, Callable, Dict, List, Set, Union


class LexicalExpander():
    '''
    A class to expand and/or normalize original lexical input terms, to
    keep back-references from generated data to corresponding original input,
    and to build consistent tokens for lexical matching.
    '''
    def __init__(
            self,
            variations_fn: Callable[[str], Set[str]],
            normalize_fn: Callable[[str], str],
            split_input_camelcase: bool = True,
            detect_emojis: bool = False,
    ):
        '''
        Initialize with the given functions.
        :param variations_fn: A function, f(t), to expand a raw input term to
          all of its variations (including itself if desired). If None, the
          default is to expand each term to itself.
        :param normalize_fn: A function to normalize a raw input term or any
          of its variations. If None, then the identity function is used.
        :param split_input_camelcase: True to split input camelcase tokens
        :param detect_emojis: True to detect emojis. If split_input_camelcase,
            then adjacent emojis will also be split; otherwise, adjacent
            emojis will appear as a single token.
        '''
        self.variations_fn = variations_fn if variations_fn else lambda x: {x}
        self.normalize_fn = normalize_fn if normalize_fn else lambda x: x
        self.split_input_camelcase = split_input_camelcase
        self.emoji_data = (
            emoji_utils.load_emoji_data()
            if detect_emojis
            else None
        )
        self.v2t = defaultdict(set)

    def __call__(self, term:Any, normalize=True) -> Set[str]:
        '''
        Get all variations of the original term.

        :param term: The term whose variations to compute.
        :param normalize: True to normalize the resulting variations.
        :return: All variations
        '''
        variations = self.variations_fn(term)
        if normalize:
            variations = {
                self.normalize_fn(v)
                for v in variations
            }
        # Add a mapping from each variation to its original term
        if variations is not None and len(variations) > 0:
            more_itertools.consume(
                map(
                    lambda v: self.v2t[v].add(term),
                    variations
                )
            )
        return variations

    def normalize(self, input_term:str) -> str:
        '''
        Normalize the given input term or variation.
        :param input_term: An input term to normalize
        :return: The normalized string of the input_term.
        '''
        return self.normalize_fn(input_term)

    def get_terms(self, variation: str) -> Set[Any]:
        '''
        Get the term ids for which the given variation was generated
        :param variation: A variation whose reference term(s) to retrieve
        :return: The set term ids for the variation or the missing_value.
        '''
        return self.v2t.get(variation, set())

    def build_first_token(
            self,
            doctext: Union[dk_doc.Text, str],
    ) -> dk_tok.Token:
        inputf = dk_tok.TextFeatures(
            doctext,
            split_camelcase=self.split_input_camelcase,
            emoji_data=self.emoji_data
        )
        return inputf.build_first_token(normalize_fn=self.normalize_fn)


class TokenMatch:
    def __init__(
            self,
            auth:dk_auth.LexicalAuthority,
            val_idx: int,
            var: str,
            token: dk_tok.Token
    ):
        self.auth = auth
        self.val_idx = val_idx
        self.var = var
        self.token = token

        self.varparts = var.split()
        self.matches = True
        self.tokens = list()
        t = token
        for v in self.varparts:
            if t is not None and v == t.norm_text:
                self.tokens.append(t)
                t = t.next_token
            else:
                self.matches = False
                break

    def __repr__(self):
        ttext = ' '.join(t.token_text for t in self.tokens)
        return f'Match_{self.tokens[0].token_num}-{self.tokens[-1].token_num}({ttext})[{self.val_idx}]'

    @property
    def next_token(self):
        next_token = None
        if self.matches:
            next_token = self.tokens[-1].next_token
        return next_token

    @property
    def matched_text(self):
        '''
        Get the matched original text.
        '''
        return self.token.input_text[
            self.tokens[0].start_pos:self.tokens[-1].end_pos
        ]

    def build_annotation(self):
        return self.auth.build_annotation(
            start_pos=self.tokens[0].start_pos,
            end_pos=self.tokens[-1].end_pos,
            entity_text=self.matched_text,
            auth_value_id=self.val_idx,
        )


class TokenAligner:
    def __init__(
            self,
            first_token: dk_tok.Token,
            authority: dk_auth.LexicalAuthority
    ):
        self.first_token = first_token
        self.auth = authority
        self.annotations = list()  # List[Dict[str, Any]]
        self._processed_idx = set()
        self._process(self.first_token)

    def _process(self, token):
        if token is not None:
            if token.token_num not in self._processed_idx:
                token_matches = self._get_token_matches(token)
                for token_match in token_matches:
                    self.annotations.append(token_match.build_annotation())
                    self._process(token_match.next_token)
            self._process(token.next_token)

    def _get_token_matches(self, token):
        token_matches = list()
        vs = self.auth.find_variations(token.norm_text, starts_with=True)
        if len(vs) > 0:
            for val_idx, var in vs.items():
                token_match = TokenMatch(self.auth, val_idx, var, token)
                if token_match.matches:
                    # mark token position(s) as matched
                    self._processed_idx.update(
                        {t.token_num for t in token_match.tokens}
                    )
                    token_matches.append(token_match)
        return token_matches


class DataframeAuthority(dk_auth.LexicalAuthority):
    '''
    A pandas dataframe-based lexical authority.
    '''
    def __init__(
            self,
            name: str,
            lexical_expander: LexicalExpander,
            authdata: dk_auth.AuthorityData,
            auth_anns_builder: dk_auth.AuthorityAnnotationsBuilder = None,
            field_groups: dk_auth.DerivedFieldGroups = None,
            anns_validator: Callable[['Authority', Dict[str,Any]], bool] = None,
            parent_auth: dk_auth.Authority = None,
    ):
        '''
        Initialize with the name, values, and associated ids of the authority;
        and with the lexical expander for authoritative values.
        
        :param name: The authority name, if different from df.columns[0]
        :param lexical_expander: The lexical expander for the values.
        :param authdata: The data for this authority
        :param auth_anns_builder: The authority annotations row builder to use
            for building annotation rows.
        :param field_groups: The derived field groups to use
        :param anns_validator: fn(auth, anns_dict_list) that returns True if
           the list of annotation row dicts are valid to be added as
           annotations for a single match or "entity".
        :param parent_auth: This authority's parent authority (if any)
        '''
        super().__init__(
            name if name else authdata.df.columns[0],
            auth_anns_builder=auth_anns_builder,
            authdata=authdata,
            field_groups=field_groups,
            anns_validator=anns_validator,
            parent_auth=parent_auth,
        )
        self.lexical_expander = lexical_expander
        self._variations = None
        self._prev_aligner = None

    @property
    def prev_aligner(self) -> TokenAligner:
        '''
        Get the token aligner created in the latest call to annotate_text.
        '''
        return self._prev_aligner

    @property
    def variations(self) -> pd.Series:
        '''
        Get all lexical variations in a series whose index has associated
        value IDs.
        :return: A pandas series with index-identified variations
        '''
        if self._variations is None:
            self._variations = self.authdata.df[self.name].apply(
                self.lexical_expander
            ).explode().dropna()
        return self._variations

    def get_id_by_variation(self, variation:str) -> Set[str]:
        '''
        Get the IDs of the value(s) associated with the given variation.
        :param variation: Variation text
        :return: The possibly empty set of associated value IDS.
        '''
        ids = set()
        for value in self.lexical_expander.get_terms(variation):
            ids.update(self.get_value_ids(value))
        return ids

    def get_variations(self, value:Any, normalize:bool = True) -> Set[Any]:
        '''
        Convenience method to compute variations for the value.
        :param value: The authority value, or term, whose variationsn to compute
        :param normalize: True to normalize the variations
        :return: The set of variations for theh value.
        '''
        return self.lexical_expander(value, normalize=normalize)

    def has_value(self, value:Any) -> bool:
        '''
        Determine whether the given value is in this authority.
        :value: A possible authority value
        :return: True if the value is a valid entity value.
        '''
        return np.any(self.authdata.df[self.name] == value)

    def get_value_ids(self, value:Any) -> Set[Any]:
        '''
        Get all IDs associated with the given value. Note that typically
        there is a single ID for any value, but this allows for inherent
        ambiguities in the authority.
        :param value: An authority value
        :return: The associated IDs or an empty set if the value is not valid.
        '''
        return set(self.authdata.lookup_values(value).index.tolist())

    def get_values_by_id(self, value_id:Any) -> Set[Any]:
        '''
        Get all values for the associated value ID. Note that typically
        there is a single value for an ID, but this allows for inherent
        ambiguities in the authority.
        
        :param value: An authority value
        :return: The associated IDs or an empty set if the value is not valid.
        '''
        return set(self.authdata.lookup_values(value_id, is_id=True)[self.name].tolist())

    def find_variations(
            self,
            variation: str,
            starts_with: bool = False,
            ends_with: bool =False,
            scope: str = 'fullmatch',
    ) -> pd.Series:
        '''
        Find all matches to the given variation.
        :param variation: The text to find; treated as a regular expression
            unless either starts_with or ends_with is True.
        :param starts_with: When True, find all terms that start with the
            variation text.
        :param ends_with: When True, find all terms that end with the variation
            text.
        :param scope: 'fullmatch' (default), 'match', or 'contains' for
            strict, less strict, and least strict matching
        :return: The matching variations as a pd.Series

        Note only the first true of starts_with, ends_with, and scope will
        be applied. If none of these are true, an full match on the pattern
        is performed.
        '''
        vs = self.variations
        if starts_with:
            vs = vs[vs.str.startswith(variation)]
        elif ends_with:
            vs = vs[vs.str.endswith(variation)]
        else:
            if scope == 'fullmatch':
                hits = vs.str.fullmatch(variation)
            elif scope == 'match':
                hits = vs.str.match(variation)
            else:
                hits = vs.str.contains(variation)
            vs = vs[hits]
        vs = vs.drop_duplicates()
        return vs

    def get_variations_df(
            self,
            variations: pd.Series,
            variations_colname: str = 'variation',
            ids_colname: str = None,
            lookup_values: bool = False,
    ) -> pd.DataFrame:
        '''
        Create a DataFrame including associated ids for each variation.
        :param variations: The variations to include in the dataframe
        :param variaions_colname: The name of the variations column
        :param id_colname: The column name for value ids.
        :param lookup_values: When True, include a self.name column
            with associated values
        '''
        if ids_colname is None:
            ids_colname = f'{self.name}_id'
        df = pd.DataFrame({
            variations_colname: variations,
            ids_colname: variations.apply(self.get_id_by_variation),
        }).explode(ids_colname)
        if lookup_values:
            df[self.name] = df[ids_colname].apply(self.get_values_by_id)
            df = df.explode(self.name)
        return df

    def add_annotations(
            self,
            doctext: dk_doc.Text,
            annotations: dk_anns.Annotations,
    ) -> dk_anns.Annotations:
        '''
        Method to do the work of finding, validating, and adding annotations.
        :param doctext: The text to process.
        :param annotations: The annotations object to add annotations to
        :return: The given or a new Annotations instance
        '''
        first_token = self.lexical_expander.build_first_token(
            doctext.text, input_id=doctext.text_id
        )
        token_aligner = TokenAligner(first_token, self)
        self._prev_aligner = token_aligner
        if self.validate_ann_dicts(token_aligner.annotations):
            annotations.add_dicts(token_aligner.annotations)
        return annotations


class CorrelatedAuthorityData(dk_auth.AuthorityData):
    '''
    Container for authoritative data containing correlated data for multiple
    "sub" authorities.
    '''
    def __init__(self, df: pd.DataFrame, name: str):
        super().__init__(df, name)
        self._authority_data = dict()

    def sub_authority_names(self) -> List[str]:
        '''
        Get the "sub" authority names.
        '''
        return None

    @abstractmethod
    def auth_values_mask(self, name: str, value_id: int) -> pd.Series:
        '''
        Identify full-authority data corresponding to this sub-value.
        :param name: The sub-authority name.
        :param value_id: The sub-authority value_id
        :return: A series representing relevant full-authority data.
        '''
        raise NotImplementedError

    @abstractmethod
    def auth_records_mask(
            self,
            record_value_ids: Dict[str, int],
            filter_mask: pd.Series = None,
    ) -> pd.Series:
        '''
        Get a series identifying records in the full authority matching
        the given records of the form {<sub-name>: <sub-value-id>}.
        :param record_value_ids: The dict of field names to value_ids
        :param filter_mask: A pre-filter limiting records to consider and/or
            building records incrementally
        :return: A series identifying where all fields exist
        '''
        raise NotImplementedError

    @abstractmethod
    def get_auth_records(self, records_mask: pd.Series) -> pd.DataFrame:
        '''
        Get the authority records identified by the mask.
        :param records_mask: A series identifying records in the full data
        :return: The records for which the mask is True.
        '''
        raise NotImplementedError

    @abstractmethod
    def combine_masks(self, mask1: pd.Series, mask2: pd.Series) -> pd.Series:
        '''
        Combine the masks if possible, returning the valid combination or None.
        :param mask1: An auth_records_mask consistent with this data
        :param mask2: Another data auth_records_mask
        :return: The combined consistent records_mask or None
        '''
        raise NotImplementedError


class MultiAuthorityData(CorrelatedAuthorityData):
    '''
    Container for authoritative data containing correlated data for multiple
    "sub" authorities composed of explicit data for each component.
    '''
    def __init__(self, df: pd.DataFrame, name: str):
        super().__init__(df, name)
        self._authority_data = dict()

    @abstractmethod
    def build_authority_data(self, name: str) -> dk_auth.AuthorityData:
        '''
        Build an authority for the named sub-authority.

        :param name: The "sub" authority name
        :return: The "sub" authority data.
        '''
        raise NotImplementedError

    @property
    def authority_data(self, name: str) -> dk_auth.AuthorityData:
        '''
        Retrieve without building the named authority data, or None
        '''
        return self._authority_data.get(name, None)

    def get_authority_data(self, name: str) -> dk_auth.AuthorityData:
        '''
        Get AuthorityData for the named "sub" authority, building if needed.

        :param name: The "sub" authority name
        :return: The "sub" authority data.
        '''
        if name not in self._authority_data:
            self._authority_data[name] = self.build_authority_data(name)
        return self._authority_data[name]

    @staticmethod
    def get_unique_vals_df(col: pd.Series, name: str) -> pd.DataFrame:
        '''
        Get a dataframe with the unique values from the column and the given
        column name.
        '''
        data = np.sort(pd.unique(col.dropna()))
        if np.issubdtype(col.dtype, np.integer):
            # IDs for an integer column are the integers themselves
            col_df = pd.DataFrame({name: data}, index=data)
        else:
            # IDs for other columns are auto-generated from 0 to n-1
            col_df = pd.DataFrame({name: data})
        return col_df

    def lookup_subauth_values(
            self,
            name: str,
            value: int,
            is_id: bool = False
    ) -> pd.DataFrame:
        '''
        Lookup "sub" authority data for the named "sub" authority value.
        :param name: The sub-authority name
        :param value: The value for the sub-authority to lookup
        :param is_id: True if value is an ID
        :return: The applicable authority dataframe rows.
        '''
        values_df = None
        authdata = self._authority_data.get(name, None)
        if authdata is not None:
            values_df = authdata.lookup_values(value, is_id=is_id)
        return values_df

    def lookup_auth_values(
            self,
            name: str,
            value: str,
    ) -> pd.DataFrame:
        '''
        Lookup original authority data for the named "sub" authority value.
        :param name: The sub-authority name
        :param value: The sub-authority value(s) (or dataframe row(s))
        :return: The original authority dataframe rows.
        '''
        return self.df[self.df[name] == value]

    def auth_values_mask(self, name: str, value_id: int) -> pd.Series:
        '''
        Identify the rows in the full authority corresponding to this sub-value.
        :param name: The sub-authority name.
        :param value_id: The sub-authority value_id
        :return: A boolean series where the field exists.
        '''
        field_values = self.lookup_subauth_values(
            name, value_id, is_id=True
        )
        return self.df[name].isin(field_values[name].tolist())

    def auth_records_mask(
            self,
            record_value_ids: Dict[str, int],
            filter_mask: pd.Series = None,
    ) -> pd.Series:
        '''
        Get a boolean series identifying records in the full authority matching
        the given records of the form {<sub-name>: <sub-value-id>}.
        :param record_value_ids: The dict of field names to value_ids
        :param filter_mask: A pre-filter limiting records to consider and/or
            building records incrementally
        :return: A boolean series where all fields exist or None
        '''
        has_fields = filter_mask
        for name, value_id in record_value_ids.items():
            has_field = self.auth_values_mask(name, value_id)
            if has_fields is None:
                has_fields = has_field
            else:
                has_fields &= has_field
        return has_fields

    def get_auth_records(self, records_mask: pd.Series) -> pd.DataFrame:
        '''
        Get the authority records identified by the mask.
        :param records_mask: A boolean series identifying records in the full df
        :return: The records/rows for which the mask is True.
        '''
        return self.df[records_mask]

    def combine_masks(self, mask1: pd.Series, mask2: pd.Series) -> pd.Series:
        '''
        Combine the masks if possible, returning the valid combination or None.
        :param mask1: An auth_records_mask consistent with this data
        :param mask2: Another data auth_records_mask
        :return: The combined consistent records_mask or None
        '''
        result = None
        if mask1 is not None and mask2 is not None:
            result = mask1 & mask2
        elif mask1 is not None:
            result = mask1
        elif mask2 is not None:
            result = mask2
        return result if np.any(result) else None


class SimpleMultiAuthorityData(MultiAuthorityData):
    '''
    Data class for pulling a single column from the multi-authority data
    as a "sub" authority.
    '''

    def build_authority_data(self, name: str) -> dk_auth.AuthorityData:
        '''
        Build an authority for the named column holding authority data.

        Note that only unique values are kept and the full dataframe's index
        will not be preserved. 

        :param name: The "sub" authority (and column) name
        :return: The "sub" authority data.
        '''
        col = self.df[name]
        col_df = self.get_unique_vals_df(col, name)
        return dk_auth.AuthorityData(col_df, name)
    

class MultiAuthorityFactory(dk_auth.AuthorityFactory):
    '''
    An factory for building a "sub" authority directly or indirectly
    from MultiAuthorityData.
    '''
    def __init__(
            self,
            auth_name: str,
            lexical_expander: LexicalExpander = None,
    ):
        '''
        :param auth_name: The name of the dataframe authority to build
        :param lexical_expander: The lexical expander to use (default=identity)
        '''
        self.auth_name = auth_name
        self._lexical_expander = lexical_expander

    def get_lexical_expander(self, name: str) -> LexicalExpander:
        '''
        Get the lexical expander for the named (column) data.
        :param name: The name of the column to expand
        :return: The appropriate lexical_expander
        '''
        if self._lexical_expander is None:
            self._lexical_expander = LexicalExpander(None, None)
        return self._lexical_expander

    def build_authority(
            self,
            name: str,
            auth_anns_builder: dk_auth.AuthorityAnnotationsBuilder,
            multiauthdata: MultiAuthorityData,
            parent_auth: dk_auth.Authority = None,
    ) -> DataframeAuthority:
        '''
        Build a DataframeAuthority.

        :param name: The name of the authority to build
        :param auth_anns_builder: The authority annotations row builder to use
            for building annotation rows.
        :param multiauthdata: The multi-authority source data
        :param parent_auth: The parent authority
        '''
        authdata = multiauthdata.get_authority_data(name)
        field_groups = None  #TODO: get from instance var set on construction?
        anns_validator = None  #TODO: get from multiauthdata?
        return DataframeAuthority(
            name,
            self.get_lexical_expander(name),
            authdata,
            field_groups=field_groups,
            anns_validator=anns_validator,
            parent_auth=parent_auth,
        )



import json
#import os
import pandas as pd
import dataknobs.utils.requests_utils as requests_utils
import dataknobs.utils.pandas_utils as pd_utils
from typing import Dict, List


def build_field_query_dict(fields, text, operator=None):
    '''
    Build an elasticsearch field query to find the text in the field(s).
    :param fields: The field (str) or fields (list[str]) to query.
    :param text: The text to find.
    :param operator: The operator to use (default if None), e.g., "AND", "OR"
    '''
    rv = None
    if isinstance(fields, str) or len(fields) == 1:  # single match
        if not isinstance(fields, str):
            fields = fields[0]
        field_dict = {'query': text}
        if operator:
            field_dict['operator'] = operator
        rv = {
            'query': {
                'match': {
                    fields: field_dict,
                }
            }
        }
    else:  # multi-match
        rv = {
            'query': {
                'multi_match': {
                    'query': text,
                    'fields': fields,
                }
            }
        }
    return rv


def build_phrase_query_dict(field, phrase, slop=0):
    '''
    Build an elasticsearch phrase query to find the phrase in the field.
    :param field: The field to query
    :param phrase: The phrase to find
    :param slop: The slop factor to use
    '''
    return {
        'query': {
            'match_phrase': {
                field: {
                    'query': phrase,
                    'slop': slop,
                }
            }
        }
    }


def build_hits_dataframe(query_result):
    '''
    Build a dataframe from an elasticsearch query result.
    '''
    hits = query_result['hits']['hits']
    dicts = [
        [hit['_source']]
        for hit in hits
    ]
    return pd_utils.dicts2df(dicts, item_id=None)


def add_batch_data(
        batchfile,
        record_generator,
        idx_name,
        source_id_fieldname="id",
        cur_id=1,
):
    '''
    Add source records from the generator to the batchfile for elasticsearch
    bulk load into the named index with record IDs starting at the given value,
    optionally adding the record ID to the source record if indicated.

    :param batchfile: The file handle to which to write the batch data
    :param record_generator: Generator for the "source" record
        dictionaries to index
    :param idx_name: Name of the elasticsearch index to hold the records
    :param source_id_fieldname: If non-empty, ensure that the source record
        also has the correct index ID in the field with this name
    :param cur_id: The id of the next source record
    :return: The id of the next source record to be added after exhausting
        the generator
    '''
    for record in record_generator:
        if source_id_fieldname:
            record[source_id_fieldname] = cur_id
        action = {
            "index": {
                "_index": idx_name,
                "_id": cur_id,
            }
        }
        print(json.dumps(action), file=batchfile)
        print(json.dumps(record), file=batchfile, flush=True)
        cur_id += 1
    return cur_id


def batchfile_record_generator(batchfile_path):
    '''
    Given the path to an elasticsearch batchfile yield each elasticsearch
    record (dict).
    :param batchfile_path: The path to the elasticdsearch batch file
    :yield: Each elasticsearch record dictionary
    '''
    with open(batchfile_path, 'r', encoding='utf-8') as f:
        for line in f:
            if not line.startswith('{"index":'):
                yield json.loads(line)


def collect_batchfile_values(batchfile_path, fieldname, default_value=''):
    '''
    Given the path to an elasticsearch batchfile and a source record fieldname,
    collect all of the values for the named field.

    :param batchfile_path: The path to the elasticsearch batchfile
    :param fieldname: The name of the source record field whose values to collect
    :param default_value: The value to use if the field doesn't exist for a record.
    :return: The list of collected values
    '''
    values = []
    for ldata in batchfile_record_generator(batchfile_path):
        values.append(ldata.get(fieldname, default_value))
    return values


def collect_batchfile_records(batchfile_path):
    '''
    Collect the batchfile records as a pandas DataFrame.
    '''
    records = []
    for record in batchfile_record_generator(batchfile_path):
        records.append(record)
    return pd.DataFrame(records)


class TableSettings:
    '''
    Container for elasticsearch table settings.
    '''

    def __init__(
            self,
            table_name,
            data_settings,
            data_mapping,
    ):
        self.name = table_name
        self.settings = data_settings
        self.mapping = data_mapping


class ElasticsearchIndex:
    '''
    Wrapper for interacting with an elasticsearch index.
    '''

    def __init__(
            self,
            request_helper,
            table_settings: List[TableSettings],
            elasticsearch_ip=None,
            elasticsearch_port=9200,
            mock_requests=None,
    ):
        self.request_helper = request_helper
        if request_helper is None:
            self.request_helper = requests_utils.RequestHelper(
                elasticsearch_ip, elasticsearch_port,
                mock_requests=mock_requests,
            )
        self.tables = table_settings
        self._init_tables()

    def _init_tables(self):
        '''
        Ensure the tables have been created and initialized.
        '''
        for table in self.tables:
            resp = self._request('get', f'{table.name}/_mapping')
            if not resp.succeeded:
                self._request('put', table.name)  # create table
                # initialize settings (for which index must be closed)
                self._request('post', f'{table.name}/_close', None)
                self._request('put', f'{table.name}/_settings', json.dumps(table.settings))
                self._request('post', f'{table.name}/_open', None)
                # initialize mappings
                self._request('put', f'{table.name}/_mapping', json.dumps(table.mapping))

    def _request(
            self,
            rtype,
            path,
            payload=None,
            params=None,
            files=None,
            response_handler=None,
            headers=None,
            verbose=False,
    ):
        return self.request_helper.request(
            rtype, path,
            payload=payload,
            params=params,
            files=files,
            response_handler=response_handler,
            headers=headers,
            verbose=verbose,
        )

    def is_up(self):
        '''
        :return: True if the elasticsearch server is up.
        '''
        return self._request('get', '_cluster/health').succeeded

    def get_cluster_health(self, verbose=False):
        '''
        :return: an requests_utils.ServerResponse instance
        '''
        return self._request('get', '_cluster/health', verbose=verbose)

    def inspect_indices(self, verbose=False):
        '''
        :return: an requests_utils.ServerResponse instance
        '''
        return self._request(
            'get',
            '_cat/indices?v&pretty',
            verbose=verbose,
            response_handler=requests_utils.plain_api_response_handler,
        )

    def purge(self, verbose=False):
        '''
        Purge all data in tables managed by this wrapper.
        '''
        resp = None
        for table in self.tables:
            resp = self.delete_table(table.name, verbose=verbose)
        self._init_tables()
        return resp

    def delete_table(self, table_name, verbose=False):
        '''
        :return: an requests_utils.ServerResponse instance
        '''
        return self._request('delete', table_name, verbose=verbose)

    ## NOT WORKING
    #def bulk_load(self, batchfilepath, verbose=False):
    #    '''
    #    Bulk load the contents of the batch file.
    #
    #    :param batchfilepath: The path to the batch file
    #    :param verbose: True to print the request response.
    #    :return: an requests_utils.ServerResponse instance
    #    '''
    #    return self._request('post-files', '_bulk', files={
    #        'batchfile': (
    #            os.path.basename(batchfilepath),
    #            open(batchfilepath, 'rb'),
    #            'Content-Type: application/x-ndjson',
    #            {'Expires': '0'}
    #        )
    #    }, verbose=verbose)

    def analyze(
            self,
            text,
            analyzer: str,
            verbose=False,
    ):
        '''
        :return: an requests_utils.ServerResponse instance
        '''
        return self._request(
            'post',
            '_analyze',
            payload=json.dumps({
                "analyzer": analyzer,
                "text": text,
            }),
            verbose=verbose
        )

    def search(
            self,
            query: Dict[str, Dict],
            table=None,
            verbose=False,
    ):
        '''
        Submit the elasticsearch search query.

        :param query: The elasticsearch search query of the form, e.g.,:
            {"query": {"match": {<field>: {"query": <text>, "operator": "AND"}}}}
        :param table: The name of the table (defaults to first table's name)
        :param verbose: True to print the request response.
        :return: an requests_utils.ServerResponse instance, resp, with resp.extra['hits_df']
            holding a dataframe representing the results if successful.
        '''
        if table is None:
            table = self.tables[0].name
        resp = self._request(
            'post', f'{table}/_search', json.dumps(query), verbose=verbose
        )
        if resp.succeeded:
            df = build_hits_dataframe(resp.result)
            resp.add_extra('hits_df', df)
        return resp
